# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# lipyphilic --- lipyphilic.readthedocs.io
#
# Released under the GNU Public Licence, v2 or any higher version
#

"""Flip-flop --- :mod:`lipyphilic.lib.flip_flop`
================================================

:Author: Paul Smith
:Year: 2021
:Copyright: GNU Public License v2

This module provides methods for finding flip-flop events in a lipid bilayer.

A flip-flop event occurs when a molecule - typically a sterol - moves from
one leaflet of a bilayer into the opposing leaflet.

The class :class:`lipyphilic.lib.flip_flop.FlipFlop` finds the frames at which
a flip-flop event begins and ends, as well as the direction of travel (upper-to-lower
or lower-to-upper). :class:`FlipFlop` can also determine whether each event was
successful (the molecule resides in the opposing leaflet for at least a given
length of time), or not (the molecule went to the midplane but returned to its
original leaflet).

See `Baral et al. (2020) <https://www.sciencedirect.com/science/article/pii/S0009308420300980>`_
for further discussion on flip-flop in lipid bilayers, including the affect on the flip-flop
rate of the buffer size used to assign molecules to the midplane of the bilayer.

Input
-----

Required:
  - *universe* : an MDAnalysis Universe object.
  - *lipid_sel* : atom selection for *all* lipids in the bilayer, including those that do not flip-flop
  - *leaflets* : leaflet membership (-1: lower leaflet, 0: midplane, 1: upper leaflet) of each lipid in the membrane at each frame
  

Output
------

  - *resindex* : residue index of a flip-flopping molecule
  - *flip_flop_start_frame* : final frame at which the molecule was present in its original leaflet
  - *flip_flop_end_frame* : first frame at which the molecule is present in the new leaflet
  - *moves_to* : direction of travel of the molecule: equal to 1 if the upper leaflet is the new lealet, equal to -1 if the lower leaflet is the new leaflet
  
Flip-flop data area returned in a :class:`numpy.ndarray`, on a "one line, one observation" basis
and can be accessed via :attr:`FlipFlop.flip_flops`::

    flip_flops = [
        [
            <resindex (0-based)>,
            <end_frame (0-based)>,
            <start_frame (0-based)>,
            <moves_to>
        ],
        ...
    ]
    
*moves_to* is equal to 1 or -1 if the molecule flip-flops into the upper or the
lower leaflet, respectively.
    
Additionaly, the success or failure of each flip-flop event is stored in the
attribute :attr:`FlipFlop.flip_flop_success`.

Example usage of :class:`FlipFlop`
----------------------------------

An MDAnalysis Universe must first be created before using :class:`FlipFlop`::

  import MDAnalysis as mda
  from lipyphilic.lib.assign_leaflets import AssignLeaflets
  from lipyphilic.lib.flip_flop import FlipFlop

  u = mda.Universe(tpr, trajectory)

Then we need to know which leaflet each lipid is in at each frame. This may be done using
:class:`lipyphilic.lib.assign_leaflets.AssignLeaflets`::

  leaflets = AssignLeaflets(
    universe=u,
    lipid_sel="name GL1 GL2 ROH"  # assuming we are using the MARTINI forcefield
    midplane_sel="name ROH,       # only cholesterol is allowed to flip-flop
    midplane_cutoff=8.0,          # buffer size for assigning molecules to the midplane
  )
  leaflets.run()

The leaflets data are stored in the :attr:`leaflets.leaflets` attribute. We can now create our
:class:`FlipFlop` object::

  flip_flop = FlipFlop(
      universe=u,
      lipid_sel="name GL1 GL2 ROH", # this must be the same as used in AssignLeaflets
      leaflets=leaflets.leaflets
  )
  
We then select which frames of the trajectory to analyse (`None` will use every
frame)::
  
  flip_flop.run(
    start=None,
    stop=None,
    step=None
  )
  
Warning
-------
    
The frames used in finding flip-flop events **must** be the same as those used for
assigning lipids to leaflets, i.e. the :attr:`start`, :attr:`stop` and :attr:`step` parameters must
be identical for :func:`AssignLeaflets.run()` and :func:`FlipFlop.run()`.


The results are then available in the :attr:`flipflop.flip_flop` attribute as a
:class:`numpy.ndarray`. Each row corresponds to an individual flip-flop event, and
the four columns correspond, respectively, to the molecule resindex,
flip-flop start frame, flip-flop end frame, and the leaflet in which the molecule
resides after the flip-flop.

We can also specify the minumum number of frames a molecule must reside in its new leaflet
for the flip-flop to be considered using. We do this using the :attr:`frame_cutoff`
parameter::

  flip_flop = FlipFlop(
      universe=u,
      lipid_sel="name GL1 GL2 ROH",
      leaflets=leaflets.leaflets,
      frame_cuotff=10,
  )

With *frame_cutoff=10*, a molecule must remain in its new leaflet for at least 10
consecutive frames for the flip-flop to be considered successful.

The class and its methods
-------------------------

.. autoclass:: FlipFlop
    :members:
    :exclude-members: run

"""
from tqdm.auto import tqdm
import numpy as np

from lipyphilic.lib import base


class FlipFlop(base.AnalysisBase):
    """Find flip-flop events in a lipid bilayer.
    """

    def __init__(self, universe,
                 lipid_sel,
                 leaflets,
                 frame_cutoff=1
                 ):
        """Set up parameters for finding flip-flop events.
        
        Parameters
        ----------
        universe : Universe
            MDAnalysis Universe object
        lipid_sel : str
            Selection string for lipids in the bilayer. This should include all lipids
            in the bilayer, including those that do not flip-flop.
        leaflets : numpy.ndarray (n_lipids,, n_frames)
            An array of leaflet membership for each lipid as each frame, in which: -1
            corresponds to the lower leaflet; 1 corresponds to the upper leaflet; and
            0 corresponds to the midplane.
        frame_cutoff : int, optional
            To be counted as a successful flip-flop, a molecule must reside in its new
            leaflet for at least 'frame_cutoff' consecutive frames. The default is `1`, in
            which case the molecule only needs to move to the opposing leaflet for a single
            frame for the flip-flop to be successful.
            
        Tip
        ---
        
        Leaflet membership can be determined using :class:`lipyphilic.lib.assign_leaflets.AssignLeaflets`.
        
        """
        super(FlipFlop, self).__init__(universe.trajectory)
        
        self.u = universe
        self.membrane = self.u.select_atoms(lipid_sel, updating=False)
        
        if (np.array(leaflets).ndim != 2) or (len(leaflets) != self.membrane.n_residues):
            raise ValueError("'leaflets' must be a 2D array of shape (n_residues, n_frames)"
                             " containing the leaflet id of each lipid at each frame."
                             )
        
        self.leaflets = np.array(leaflets)
        
        if frame_cutoff < 1:
            raise ValueError("'frame_cutoff' must be greater than or equal to 1")
           
        self.frame_cutoff = frame_cutoff
        
        self.flip_flops = None
        self.flip_flop_success = None
        
    def _setup_frames(self, trajectory, start=None, stop=None, step=None):
        """
        Pass a Reader object and define the desired iteration pattern
        through the trajectory

        Parameters
        ----------
        trajectory : MDAnalysis.Reader
            A trajectory Reader
        start : int, optional
            start frame of analysis
        stop : int, optional
            stop frame of analysis
        step : int, optional
            number of frames to skip between each analysed frame
        
        """
        self._trajectory = trajectory
        start, stop, step = trajectory.check_slice_indices(start, stop, step)
        self.start = start
        self.stop = stop
        self.step = step
        
        n_frames = len(range(start, stop, step))
        if self.leaflets.shape[1] != n_frames:
            raise ValueError("The frames to analyse must be identical to those used "
                             "in assigning lipids to leaflets."
                             )
        
        self.n_frames = n_frames
        self.frames = np.array([ts.frame for ts in trajectory[start:stop:step]], dtype=int)
          
    def _prepare(self):
        
        # Output array
        self.flip_flops = [[], [], [], []]
        self.flip_flop_success = []
        
    def _single_frame(self):
        
        # Skip if the molecule never changes leaflet
        if np.min(np.diff(self._residue_leaflets)) == np.max(np.diff(self._residue_leaflets)) == 0:
            return None
        
        # Check when the molecule leaves the upper leaflet for more than `frame_cutoff` frames
        # If it has left for at least this many frames, there is a chance it has
        # flip-flopped to the opposing leaflet
        leaflet = 1
        residue_leaflet_indices = np.nonzero(self._residue_leaflets == leaflet)[0]
        gaps = np.diff(residue_leaflet_indices) > self.frame_cutoff

        upper_begins = np.insert(residue_leaflet_indices[1:][gaps], 0, residue_leaflet_indices[0])
        upper_ends = np.append(residue_leaflet_indices[:-1][gaps], residue_leaflet_indices[-1])
        
        # Check when the molecule leaves the lower leaflet for more than `frame_cutoff` frames
        leaflet = -1
        residue_leaflet_indices = np.nonzero(self._residue_leaflets == leaflet)[0]
        gaps = np.diff(residue_leaflet_indices) > self.frame_cutoff

        lower_begins = np.insert(residue_leaflet_indices[1:][gaps], 0, residue_leaflet_indices[0])
        lower_ends = np.append(residue_leaflet_indices[:-1][gaps], residue_leaflet_indices[-1])

        # Combine beginnings and ending of leaflet membership
        begins = np.array(list(lower_begins) + list(upper_begins))
        ends = np.array(list(lower_ends) + list(upper_ends))
        moves_to = np.array([-1] * len(lower_begins) + [1] * len(upper_begins))

        sort = np.argsort(begins)
        begins = begins[sort]
        ends = ends[sort]
        moves_to = moves_to[sort]

        # To be considered to have flip-flopped, the molecule must remain in the leaflet for at least `frame_cutoff` frames
        keep = (ends - begins) >= (self.frame_cutoff - 1)  # if end==begin, the molecule was there for 1 frame not 0 frames
        begins = begins[keep]
        ends = ends[keep]
        moves_to = moves_to[keep]

        # Store data for when leaflet membership ends, begins, and which leaflet it has moved to
        resindex = self.membrane[self._residue_index].resindex
        
        self.flip_flops[0].extend(np.full_like(begins[1:], fill_value=resindex))
        self.flip_flops[1].extend(self.frames[ends[:-1]])
        self.flip_flops[2].extend(self.frames[begins[1:]])
        self.flip_flops[3].extend(moves_to[1:])
        
        # Check whether the flip-flop was a success or if it moved back to its
        # original leaflet
        success = np.full_like(moves_to[1:], fill_value="Success", dtype="<U10")
        success[1:][np.diff(moves_to[1:]) == 0] = "Fail"
        self.flip_flop_success.extend(success)
        
    def _conclude(self):
    
        self.flip_flops = np.asarray(self.flip_flops).T
        self.flip_flop_success = np.asarray(self.flip_flop_success)
    
    def run(self, start=None, stop=None, step=None):
        """Perform the calculation

        Parameters
        ----------
        start : int, optional
            start frame of analysis
        stop : int, optional
            stop frame of analysis
        step : int, optional
            number of frames to skip between each analysed frame
        """
        
        self._setup_frames(self._trajectory, start, stop, step)
        self._prepare()
        
        for residue_index, residue_leaflets in tqdm(enumerate(
                self.leaflets), total=self.membrane.n_residues):
            
            self._residue_index = residue_index
            self._residue_leaflets = residue_leaflets
            self._single_frame()
            
        self._conclude()
        return self



"""
Simple systems created for testing lipyphilic
=============================================

"""
__all__ = [
    "HEX_LAT",
    "HEX_LAT_BUMP",
    "HEX_LAT_BUMP_MID_MOL",
    "HEX_LAT_BUMP_MID_ATOM",
    "HEX_LAT_OVERLAP",
    "ONE_CHOL",
    "ONE_CHOL_TRAJ",
]

from pkg_resources import resource_filename

# Two hexagonal lattices to represent upper and lower leaflets, separated in
# z by 20 Angstrom
# Alternating residues of LIPID and CHOL
# Two atoms per molecule
# Atoms are equally spaced
HEX_LAT = resource_filename(__name__,
                            "pdbs/HexGrid-2AtomsPerLipid.pdb")
# As immediately above but the central part of the lattices translated in
# z by +5 Angstrom
HEX_LAT_BUMP = resource_filename(__name__,
                                 "pdbs/HexGrid-2AtomsPerLipid-Undulating.pdb")
# As immediately above but one molecule (CHOL with resid 78) has been
# moved to the midplane
HEX_LAT_BUMP_MID_MOL = resource_filename(__name__,
                                         "pdbs/HexGrid-2AtomsPerLipid-Undulating-Midplane-Molecule.pdb")
# Same as HEX_LAT_BUMP but one atom of one molecule (CHOL with resid 78) has been
# moved to the midplane
HEX_LAT_BUMP_MID_ATOM = resource_filename(__name__,
                                          "pdbs/HexGrid-2AtomsPerLipid-Undulating-Midplane-Atom.pdb")
# Same as HEX_LAT but the first two atoms are overalpping (have to exact same position)
HEX_LAT_OVERLAP = resource_filename(__name__,
                                    "pdbs/HexGrid-2AtomsPerLipid-OverlappingAtoms.pdb")

# A single coarse-grained cholesterol molecule
ONE_CHOL = resource_filename(__name__,
                             "pdbs/Chol-Flip-Flop.pdb")
# A trajectory of 25 frames of the above cholesterol molecule
ONE_CHOL_TRAJ = resource_filename(__name__,
                                  "xtcs/Chol-Flip-Flop.xtc")




import pytest
import numpy as np
import MDAnalysis

from numpy.testing import assert_array_almost_equal

from lipyphilic._simple_systems.simple_systems import (
    HEX_LAT, HEX_LAT_BUMP_MID_MOL, HEX_LAT_OVERLAP)
from lipyphilic.lib.area_per_lipid import AreaPerLipid
 
 
class TestAreaPerLipid:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)

    # first 50 residues are in the upper leaflet (1)
    # final 50 residues are in the upper leaflet (-1)
    kwargs = {
        'lipid_sel': 'name L C',
        'leaflets': np.array([[1]] * 50 + [[-1]] * 50)
    }
    
    @pytest.fixture(scope='class')
    def areas(self, universe):
        areas = AreaPerLipid(universe, **self.kwargs)
        areas.run()
        return areas

    def test_area_per_lipid(self, areas):
    
        reference = {
            'n_residues': 100,
            'n_frames': 1,
            'area': 200  # all lipids have the same area per lipid as they are on a hexagonal lattice
        }
    
        assert areas.areas.shape == (reference['n_residues'], reference['n_frames'])
        assert_array_almost_equal(areas.areas, 200.0, decimal=8)
        
        
class TestAreaPerLipidOverlapping:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT_OVERLAP)

    # first 50 residues are in the upper leaflet (1)
    # final 50 residues are in the upper leaflet (-1)
    kwargs = {
        'lipid_sel': 'name L C',
        'leaflets': np.array([[1]] * 50 + [[-1]] * 50)
    }
    
    @pytest.fixture(scope='class')
    def areas(self, universe):
        areas = AreaPerLipid(universe, **self.kwargs)
        areas.run()
        return areas
    
    def test_area_per_lipid(self, areas):
        
        # only the areas of 6 residues should be affected by these two overlapping atoms
        assert np.isclose(areas.areas, 200).sum() == 94


class TestAreaPerLipidMidplaneMol:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT_BUMP_MID_MOL)

    # first 50 residues are in the upper leaflet (1)
    # final 50 residues are in the upper leaflet (-1), except:
    # CHOL78 is in the midplane
    leaflets = np.array([[1]] * 50 + [[-1]] * 50)
    leaflets[78] = 0
    kwargs = {
        'lipid_sel': 'name L C',
        'leaflets': leaflets
    }
    
    @pytest.fixture(scope='class')
    def areas(self, universe):
        areas = AreaPerLipid(universe, **self.kwargs)
        areas.run()
        return areas

    def test_area_per_lipid(self, areas):
    
        reference = {
            'n_residues': 100,
            'n_frames': 1,
            'min_area': 200  # all lipids have at least this area
        }
    
        assert areas.areas.shape == (reference['n_residues'], reference['n_frames'])
        assert_array_almost_equal(np.nanmin(areas.areas), 200.0, decimal=8)
        assert np.isnan(areas.areas[78])
        assert sum(np.isnan(areas.areas)) == 1


class TestAreaPerLipidExceptions:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)

    def test_Exceptions(self, universe):
            
        match = "'leaflets' must either be a 1D array containing non-changing "
        with pytest.raises(ValueError, match=match):
            AreaPerLipid(
                universe=universe,
                lipid_sel="name L C",
                leaflets=None
            )
        
        with pytest.raises(ValueError, match=match):
            AreaPerLipid(
                universe=universe,
                lipid_sel="name L C",
                leaflets=np.array([[[], []], [[], []]])  # cannot pass a 3D array
            )
        
        match = ("The shape of 'leaflets' must be \\(n_residues,\\), but 'lipid_sel' "
                 "generates an AtomGroup with 100 residues"
                 " and 'leaflets' has shape \\(99, 1\\).")
        with pytest.raises(ValueError, match=match):
            AreaPerLipid(
                universe=universe,
                lipid_sel="name L C",
                leaflets=np.array([[1]] * 50 + [[-1]] * 49)  # one residue too few
            )
            
        match = ("The frames to analyse must be identical to those used "
                 "in assigning lipids to leaflets.")
        with pytest.raises(ValueError, match=match):
            areas = AreaPerLipid(
                universe=universe,
                lipid_sel="name L C",
                leaflets=np.array([[1, 1]] * 50 + [[-1, -1]] * 50)  # leaflets has two frames, apl one
            )
            areas.run()



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# lipyphilic --- lipyphilic.readthedocs.io
#
# Released under the GNU Public Licence, v2 or any higher version
#

"""Area per lipid --- :mod:`lipyphilic.lib.area_per_lipid`
==========================================================

:Author: Paul Smith
:Year: 2021
:Copyright: GNU Public License v2

This module provides methods for calculating the area per lipid in a bilayer.

The class :class:`lipyphilic.lib.area_per_lipid.AreaPerLipid` calculates the
area of each lipid via a 2D Voronoi tessellation of atomic positions. See
`Lukat et al. (2013) <https://pubs.acs.org/doi/full/10.1021/ci400172g>`_ for
a description of calculating the area per lipid via Voronoi tessellations.

This class uses `Freud <https://freud.readthedocs.io/en/stable/index.html#>`_
for performing the Voronoi tessellations from which the area per lipid is
calculated.

Input
------

Required:
  - *universe* : an MDAnalysis Universe object.
  - *lipid_sel* : atom selection for lipids in the bilayer. These atoms will be used to perform the Voronoi tessellation.
  - *leaflets* : leaflet membership (-1: lower leaflet, 0: midplane, 1: upper leaflet) of each lipid in the membrane.
  

Output
------

  - *area* : area per lipid of each lipid as each frame
  
Area data are returned in a :class:`numpy.ndarray`, where each row corresponds
to an individual lipid and each columns corresponds to an individual frame. The
results are accessible via the :attr:`AssignLeaflets.leaflets` attribute.

Area data are returned in a :class:`numpy.ndarray`, where each row corresponds
to an individual lipid and each column corresponds to an individual frame, i.e.
areas[i, j] refers to the area of lipid *i* at frame *j*. The results are
accessible via the :attr:`AreaPerLipid.areas` attribute.

Note
----

No area can be calculated for molecules that are in the midplane,
i.e. those for which `leaflets==0`. These molecules will have `NaN` values
in the results array for the frames at which they are in the midplane.


Example usage of :class:`AreaPerLipid`
--------------------------------------

An MDAnalysis Universe must first be created before using AreaPerLipid::

  import MDAnalysis as mda
  from lipyphilic.lib.assign_leaflets import AssignLeaflets

  u = mda.Universe(tpr, trajectory)

Then we need to know which leaflet each lipid is in at each frame. This may be done using
:class:`lipyphilic.lib.assign_leaflets.AssignLeaflets`::

  leaflets = AssignLeaflets(
    universe=u,
    lipid_sel="name GL1 GL2 ROH" # assuming we are using the MARTINI forcefield
  )
  leaflets.run()

The leaflets data are stored in the :attr:`leaflets.leaflets` attribute. We can now create our
AreaPerLipid object::

  areas = AreaPerLipid(
      universe=u,
      lipid_sel="name GL1 GL2 ROH",
      leaflets=leaflets.leaflets
  )
  
The above will use GL1 and GL2 beads to calculate the area of each phospholipid, and the
ROH bead to calculate the area of each sterol.

We then select which frames of the trajectory to analyse (`None` will use every
frame) and choose to display a progress bar (`verbose=True`)::
  
  areas.run(
    start=None,
    stop=None,
    step=None,
    verbose=True
  )
  
Warning
-------
    
The frames used in calculating the area per lipid must be the same as those used for
assigning lipids to leaflets, i.e. the :attr:`start`, :attr:`stop` and :attr:`step` parameters must
be identical for :func:`AssignLeaflets.run()` and :func:`AreaPerLipid.run()`.


The results are then available in the :attr:`areas.areas` attribute as a
:class:`numpy.ndarray`. Each row corresponds to an individual lipid and each column
to an individual frame, i.e `areas.areas[i, j]` contains the area of lipid *i*
at frame *j*.


The class and its methods
-------------------------

.. autoclass:: AreaPerLipid
    :members:

"""
import numpy as np
import freud.locality

from lipyphilic.lib import base


class AreaPerLipid(base.AnalysisBase):
    """Calculate the area of lipids in each leaflet of a bilayer.
    """

    def __init__(self, universe,
                 lipid_sel,
                 leaflets
                 ):
        """Set up parameters for calculating areas.
        
        Parameters
        ----------
        universe : Universe
            MDAnalysis Universe object
        lipid_sel : str
            Selection string for lipids in the bilayer. Typically, in all-atom
            simulations, one atom per sterol and three atoms per non-sterol lipid
            would be used. In coarse-grained simulations, one bead per sterol and
            two beads per non-sterol lipid would typically be used.
        leaflets : numpy.ndarray (n_lipids,)
            An array of leaflet membership in which: -1 corresponds to the lower leaflet;
            1 corresponds to the upper leaflet; and 0 corresponds to the midplane.
            If the array is 1D and of shape (n_lipids), each lipid is taken to
            remain in the same leaflet over the trajectory. If the array is 2D and
            of shape (n_lipids, n_frames), the leaflet to which each lipid is
            assisgned at each frame will be taken into account when calculating
            the area per lipid.
            
        Tip
        ---
        
        Leaflet membership can be determined using :class:`lipyphilic.lib.assign_leaflets.AssignLeaflets`.
        
        """
        super(AreaPerLipid, self).__init__(universe.trajectory)

        self.u = universe
        self.membrane = self.u.select_atoms(lipid_sel, updating=False)
        
        if np.array(leaflets).ndim not in [1, 2]:
            raise ValueError("'leaflets' must either be a 1D array containing non-changing "
                             "leaflet ids of each lipid, or a 2D array of shape (n_residues, n_frames)"
                             " containing the leaflet id of each lipid at each frame."
                             )

        if len(leaflets) != self.membrane.n_residues:
            raise ValueError("The shape of 'leaflets' must be (n_residues,), but 'lipid_sel' "
                             f"generates an AtomGroup with {self.membrane.n_residues} residues"
                             f" and 'leaflets' has shape {leaflets.shape}."
                             )
        
        self.leaflets = np.array(leaflets)
        
        # lipid species in the membrane
        self._lipid_species = np.unique(self.membrane.resnames)
        # number of each lipid species in the membrane
        num_lipids = {lipid: sum(self.membrane.residues.resnames == lipid) for lipid in self._lipid_species}
        # number of atoms (seeds) used in the Voronoi tessellation per molecule for each species
        self._num_seeds = {
            lipid: sum(self.membrane.resnames == lipid) // num_lipids[lipid] for lipid in self._lipid_species
        }
        
        self.areas = None
          
    def _prepare(self):
        
        if (self.leaflets.ndim == 2) and (self.leaflets.shape[1] != self.n_frames):
            raise ValueError("The frames to analyse must be identical to those used "
                             "in assigning lipids to leaflets."
                             )
        
        # Output array
        self.areas = np.full(
            (self.membrane.n_residues, self.n_frames),
            fill_value=np.NaN,
            dtype=float
        )
        
    def _single_frame(self):
        
        # Atoms must be wrapped before creating a lateral grid of the membrane
        self.membrane.wrap(inplace=True)
        frame_leaflets = self.leaflets[:, self._frame_index] if self.leaflets.ndim == 2 else self.leaflets
        
        # Calculate area per lipid for the lower (-1) and upper (1) leaflets
        # Areas cannot be calculated for midplane (0) molecules.
        for leaflet_sign in [-1, 1]:
            
            # freud.order.Voronoi requires z positions set to 0
            leaflet = self.membrane.residues[frame_leaflets == leaflet_sign].atoms
            atoms = leaflet.atoms.intersection(self.membrane)
            pos = atoms.positions
            pos[:, 2] = 0
            
            # Check whether any atoms are overlapping in the xy-plane
            self._remove_overlapping(positions=pos)

            # Voronoi tessellation to get area per atom
            areas = self._get_atom_areas(positions=pos)

            # Calculae area per lipid in the current leaflet
            # by considering the contribution of each
            # atom of a given lipid
            self._get_area_per_lipid(
                atoms=atoms,
                atom_areas=areas
            )
            
    def _remove_overlapping(self, positions):
        """Ensure no two atoms are overlapping in the xy plane.
        
        Given an Nx3 array of atomic positions, make minor adjustments to xy positions
        if any pair of xy coordinates are identical.
        
        If atoms are overlapping in xy, Freud will complain when attempting to perform the
        Voronoi tessellation.
        
        Parameters
        ----------
        positions : numpy ndarray
            Array of shape (n_atoms, 3) containing atomic coordinates.
        
        Returns
        -------
        None
            The positions are modified in place.
        
        """
        
        # Check whether any atoms are overlapping in the xy-plane
        # This may be an issue in CG sims with cholesteorl flip-flop
        # but is unlikely to be so in all-atom sims
        _, indices, counts = np.unique(
            positions, return_index=True, return_counts=True, axis=0
        )
        
        # If so, add a small distance between the two atoms (1e-3 A)
        # in the x dimension
        if max(counts > 1):
            for duplicate_index in indices[counts > 1]:
                positions[duplicate_index, 0] += 0.001
                
        return None

    def _get_atom_areas(self, positions):
        """Calculate area per atom.
        
        Given xy coordinates of atomic positions, perform a Voronoi
        tessellation and return the area in xy occupied by each Voronoi cell.
        
        Parameters
        ----------
        positions : numpy ndarray
            Array of shape (n_atoms, 3) containing atomic coordinates.
        
        Returns
        -------
        areas : numpy ndarray
            Array of shape (n_atoms) containing the lateral area per atom.
        
        """
        
        voro = freud.locality.Voronoi()
        areas = voro.compute(
            system=(
                {
                    "Lx": self._ts.dimensions[0],
                    "Ly": self._ts.dimensions[1],
                    "dimensions": 2
                },
                positions
            )
        ).volumes
        
        return areas
    
    def _get_area_per_lipid(self, atoms, atom_areas):
        """Calclate the area per lipid given the areas of every Voronoi cell in a tessellation.
        
        This involves summing contributions from each atom of a given lipid.
        
        Parameters
        ----------
        atoms : MDAnalysis AtomGroup
            AtomGroup for which a 2D Voronoi tessellation was performed
        atom_areas : numpy ndarray
            Array of areas of each atom in the 2D Voronoi tessellation
        
        Returns
        -------
        None
            The lipid areas are modified in place.
        """
        
        for species in self._lipid_species:

            species_indices = atoms.resnames == species
            
            # We need to sum the area contribution of each cell for a given lipid
            species_apl = atom_areas[species_indices]
            species_atoms = atoms[species_indices]
            species_apl = np.sum(
                species_apl.reshape(species_atoms.n_residues, self._num_seeds[species]),
                axis=1
            )

            # store apl for current lipid species
            species_resindices = np.in1d(
                self.membrane.residues.resindices,
                species_atoms.residues.resindices,
                assume_unique=True
            )
            
            self.areas[species_resindices, self._frame_index] = species_apl
            
        return None



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Analysis building blocks --- :mod:`MDAnalysis.analysis.base`
============================================================

A collection of useful building blocks for creating Analysis
classes.

"""
import inspect
import itertools
import logging

import numpy as np
from MDAnalysis import coordinates
from MDAnalysis.core.groups import AtomGroup
from MDAnalysis.lib.log import ProgressBar

logger = logging.getLogger(__name__)


class AnalysisBase(object):
    """Base class for defining multi frame analysis

    The class it is designed as a template for creating multiframe analyses.
    This class will automatically take care of setting up the trajectory
    reader for iterating, and it offers to show a progress meter.

    To define a new Analysis, `AnalysisBase` needs to be subclassed
    `_single_frame` must be defined. It is also possible to define
    `_prepare` and `_conclude` for pre and post processing. See the example
    below.

    .. code-block:: python

       class NewAnalysis(AnalysisBase):
           def __init__(self, atomgroup, parameter, **kwargs):
               super(NewAnalysis, self).__init__(atomgroup.universe.trajectory,
                                                 **kwargs)
               self._parameter = parameter
               self._ag = atomgroup

           def _prepare(self):
               # OPTIONAL
               # Called before iteration on the trajectory has begun.
               # Data structures can be set up at this time
               self.result = []

           def _single_frame(self):
               # REQUIRED
               # Called after the trajectory is moved onto each new frame.
               # store result of `some_function` for a single frame
               self.result.append(some_function(self._ag, self._parameter))

           def _conclude(self):
               # OPTIONAL
               # Called once iteration on the trajectory is finished.
               # Apply normalisation and averaging to results here.
               self.result = np.asarray(self.result) / np.sum(self.result)

    Afterwards the new analysis can be run like this.

    .. code-block:: python

       na = NewAnalysis(u.select_atoms('name CA'), 35).run(start=10, stop=20)
       print(na.result)

    Attributes
    ----------
    times: np.ndarray
        array of Timestep times. Only exists after calling run()
    frames: np.ndarray
        array of Timestep frame indices. Only exists after calling run()

    """

    def __init__(self, trajectory, verbose=False, **kwargs):
        """
        Parameters
        ----------
        trajectory : mda.Reader
            A trajectory Reader
        verbose : bool, optional
           Turn on more logging and debugging, default ``False``


        .. versionchanged:: 1.0.0
           Support for setting ``start``, ``stop``, and ``step`` has been
           removed. These should now be directly passed to
           :meth:`AnalysisBase.run`.
        """
        self._trajectory = trajectory
        self._verbose = verbose

    def _setup_frames(self, trajectory, start=None, stop=None, step=None):
        """
        Pass a Reader object and define the desired iteration pattern
        through the trajectory

        Parameters
        ----------
        trajectory : mda.Reader
            A trajectory Reader
        start : int, optional
            start frame of analysis
        stop : int, optional
            stop frame of analysis
        step : int, optional
            number of frames to skip between each analysed frame


        .. versionchanged:: 1.0.0
            Added .frames and .times arrays as attributes

        """
        self._trajectory = trajectory
        start, stop, step = trajectory.check_slice_indices(start, stop, step)
        self.start = start
        self.stop = stop
        self.step = step
        self.n_frames = len(range(start, stop, step))
        self.frames = np.zeros(self.n_frames, dtype=int)
        self.times = np.zeros(self.n_frames)

    def _single_frame(self):
        """Calculate data from a single frame of trajectory

        Don't worry about normalising, just deal with a single frame.
        """
        raise NotImplementedError("Only implemented in child classes")

    def _prepare(self):
        """Set things up before the analysis loop begins"""
        pass  # pylint: disable=unnecessary-pass

    def _conclude(self):
        """Finalise the results you've gathered.

        Called at the end of the run() method to finish everything up.
        """
        pass  # pylint: disable=unnecessary-pass

    def run(self, start=None, stop=None, step=None, verbose=None):
        """Perform the calculation

        Parameters
        ----------
        start : int, optional
            start frame of analysis
        stop : int, optional
            stop frame of analysis
        step : int, optional
            number of frames to skip between each analysed frame
        verbose : bool, optional
            Turn on verbosity
        """
        logger.info("Choosing frames to analyze")
        # if verbose unchanged, use class default
        verbose = getattr(self, '_verbose',
                          False) if verbose is None else verbose

        self._setup_frames(self._trajectory, start, stop, step)
        logger.info("Starting preparation")
        self._prepare()
        for i, ts in enumerate(ProgressBar(
                self._trajectory[self.start:self.stop:self.step],
                verbose=verbose)):
            self._frame_index = i
            self._ts = ts
            self.frames[i] = ts.frame
            self.times[i] = ts.time
            # logger.info("--> Doing frame {} of {}".format(i+1, self.n_frames))
            self._single_frame()
        logger.info("Finishing up")
        self._conclude()
        return self


class AnalysisFromFunction(AnalysisBase):
    """
    Create an analysis from a function working on AtomGroups

    Attributes
    ----------
    results : ndarray
        results of calculation are stored after call to ``run``

    Example
    -------
    >>> def rotation_matrix(mobile, ref):
    >>>     return mda.analysis.align.rotation_matrix(mobile, ref)[0]

    >>> rot = AnalysisFromFunction(rotation_matrix, trajectory, mobile, ref).run()
    >>> print(rot.results)

    Raises
    ------
    ValueError : if ``function`` has the same kwargs as ``BaseAnalysis``
    """

    def __init__(self, function, trajectory=None, *args, **kwargs):
        """Parameters
        ----------
        function : callable
            function to evaluate at each frame
        trajectory : mda.coordinates.Reader (optional)
            trajectory to iterate over. If ``None`` the first AtomGroup found in
            args and kwargs is used as a source for the trajectory.
        *args : list
            arguments for ``function``
        **kwargs : dict
            arguments for ``function`` and ``AnalysisBase``

        .. versionchanged:: 1.0.0
           Support for directly passing the ``start``, ``stop``, and ``step``
           arguments has been removed. These should instead be passed
           to :meth:`AnalysisFromFunction.run`.

        """
        if (trajectory is not None) and (not isinstance(
                trajectory, coordinates.base.ProtoReader)):
            args = (trajectory,) + args
            trajectory = None

        if trajectory is None:
            # all possible places to find trajectory
            for arg in itertools.chain(args, kwargs.values()):
                if isinstance(arg, AtomGroup):
                    trajectory = arg.universe.trajectory
                    break

        if trajectory is None:
            raise ValueError("Couldn't find a trajectory")

        self.function = function
        self.args = args

        self.kwargs = kwargs

        super(AnalysisFromFunction, self).__init__(trajectory)

    def _prepare(self):
        self.results = []

    def _single_frame(self):
        self.results.append(self.function(*self.args, **self.kwargs))

    def _conclude(self):
        self.results = np.asarray(self.results)


def analysis_class(function):
    """
    Transform a function operating on a single frame to an analysis class

    For an usage in a library we recommend the following style:

    >>> def rotation_matrix(mobile, ref):
    >>>     return mda.analysis.align.rotation_matrix(mobile, ref)[0]
    >>> RotationMatrix = analysis_class(rotation_matrix)

    It can also be used as a decorator:

    >>> @analysis_class
    >>> def RotationMatrix(mobile, ref):
    >>>     return mda.analysis.align.rotation_matrix(mobile, ref)[0]

    >>> rot = RotationMatrix(u.trajectory, mobile, ref).run(step=2)
    >>> print(rot.results)
    """

    class WrapperClass(AnalysisFromFunction):
        def __init__(self, trajectory=None, *args, **kwargs):
            super(WrapperClass, self).__init__(function, trajectory,
                                               *args, **kwargs)

    return WrapperClass


def _filter_baseanalysis_kwargs(function, kwargs):
    """
    create two dictionaries with kwargs separated for function and AnalysisBase

    Parameters
    ----------
    function : callable
        function to be called
    kwargs : dict
        keyword argument dictionary

    Returns
    -------
    base_args : dict
        dictionary of AnalysisBase kwargs
    kwargs : dict
        kwargs without AnalysisBase kwargs

    Raises
    ------
    ValueError : if ``function`` has the same kwargs as ``BaseAnalysis``
    """
    try:
        # pylint: disable=deprecated-method
        base_argspec = inspect.getfullargspec(AnalysisBase.__init__)
    except AttributeError:
        # pylint: disable=deprecated-method
        base_argspec = inspect.getargspec(AnalysisBase.__init__)

    n_base_defaults = len(base_argspec.defaults)
    base_kwargs = {name: val
                   for name, val in zip(base_argspec.args[-n_base_defaults:],
                                        base_argspec.defaults)}

    try:
        # pylint: disable=deprecated-method
        argspec = inspect.getfullargspec(function)
    except AttributeError:
        # pylint: disable=deprecated-method
        argspec = inspect.getargspec(function)

    for base_kw in base_kwargs.keys():
        if base_kw in argspec.args:
            raise ValueError(
                "argument name '{}' clashes with AnalysisBase argument."
                "Now allowed are: {}".format(base_kw, base_kwargs.keys()))

    base_args = {}
    for argname, default in base_kwargs.items():
        base_args[argname] = kwargs.pop(argname, default)

    return base_args, kwargs





import pytest
import numpy as np
from numpy.testing import assert_array_equal
import MDAnalysis
from MDAnalysis.exceptions import NoDataError

from lipyphilic._simple_systems.simple_systems import HEX_LAT
from lipyphilic.lib.neighbours import Neighbours
 
 
class TestNeighbours:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)

    kwargs = {
        'lipid_sel': 'name L C'
    }
    
    def test_neighbours_cutoff12(self, universe):
    
        neighbours = Neighbours(universe, **self.kwargs, cutoff=12)
        neighbours.run()
    
        # it's a hexagonal lattice
        # with cutoff=12, every lipid should have 6 neighbours
        reference = {
            'n_residues': 100,  # there are 200 atoms but 100 lipids in total
            'n_neighbours': 6,
        }
    
        assert neighbours.neighbours.shape == (reference['n_residues'], reference['n_residues'])
        assert (np.sum(neighbours.neighbours.toarray(), axis=0) == reference['n_neighbours']).all()
        
    def test_neighbours_cutoff10(self, universe):
        
        neighbours = Neighbours(universe, **self.kwargs, cutoff=10)
        neighbours.run()
    
        # it's a hexagonal lattice, but each hexagon is irregular (two sides longer than the other 4)
        # with cutoff=10, every lipid should have 2 neighbours
        reference = {
            'n_residues': 100,  # there are 200 atoms but 100 lipids in total
            'n_neighbours': 2,
        }
    
        assert neighbours.neighbours.shape == (reference['n_residues'], reference['n_residues'])
        assert (np.sum(neighbours.neighbours.toarray(), axis=0) == reference['n_neighbours']).all()
        
    def test_subset_lipids(self, universe):
        
        neighbours = Neighbours(universe, lipid_sel="name C", cutoff=10)
        neighbours.run()
    
        # it's a hexagonal lattice, but each hexagon is irregular (two sides longer than the other 4)
        # with cutoff=10, every lipid should have 2 neighbours
        reference = {
            'n_residues': 50,
            'n_neighbours': np.array(
                [
                    0, 0, 1, 0, 1, 0, 0, 1, 0, 1,
                    0, 0, 1, 0, 1, 0, 0, 1, 0, 1,
                    0, 0, 1, 0, 1, 0, 0, 1, 0, 1,
                    0, 0, 1, 0, 1, 0, 0, 1, 0, 1,
                    0, 0, 1, 0, 1, 0, 0, 1, 0, 1
                ]
            )
        }
    
        assert neighbours.neighbours.shape == (reference['n_residues'], reference['n_residues'])
        assert_array_equal(np.sum(neighbours.neighbours.toarray(), axis=0), reference['n_neighbours'])


class TestNeighboursExceptions:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)

    def test_Exceptions(self, universe):
            
        match = "'cutoff' must be greater than 0"
        with pytest.raises(ValueError, match=match):
            Neighbours(
                universe=universe,
                lipid_sel="name L C",
                cutoff=-1
            )


class TestNeighboursCount:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)

    kwargs = {
        'lipid_sel': 'name L C',
        'cutoff': 12.0
    }
    
    @pytest.fixture(scope='class')
    def neighbours(self, universe):
        neighbours = Neighbours(universe, **self.kwargs)
        neighbours.run()
        return neighbours
    
    # the sequence of n_CHOL_neighbours and n_LIPID_neighbours
    # arise because - even though the atoms are arranged on a
    # hexagonal lattice - each residue has two atoms
    @pytest.fixture(scope='class')
    def reference(self):
        
        reference = {
            'n_residues': 100,
            'n_frames': 1,
            'n_columns': 6,
            'n_neighbours': 6,
            'n_CHOL_neighbours': np.array(
                [
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3,
                    3, 2, 4, 2, 5, 1, 4, 2, 4, 3
                ]
            )
        }
        reference['n_LIPID_neighbours'] = 6 - reference['n_CHOL_neighbours']
        
        return reference
    
    def test_count_neighbours(self, neighbours, reference):
        
        counts = neighbours.count_neighbours()
        
        assert_array_equal(counts.shape, (reference['n_residues'] * reference['n_frames'], reference['n_columns']))
        assert_array_equal(counts.Frame, np.zeros(reference['n_residues']))
        assert (counts.Total == reference['n_neighbours']).all()
        assert_array_equal(counts.nCHOL, reference['n_CHOL_neighbours'])
        assert_array_equal(counts.nLIPI, reference['n_LIPID_neighbours'])
    
    def test_count_neighbours_count_by(self, neighbours, reference):
        
        # make every LIPI take the value 0
        # and every CHOL take the value 1
        count_by = np.zeros((reference['n_residues'], reference['n_frames']), dtype=np.int8)
        count_by[1::2] = 1
        
        counts = neighbours.count_neighbours(count_by=count_by)
        
        assert_array_equal(counts.shape, (reference['n_residues'] * reference['n_frames'], reference['n_columns']))
        assert_array_equal(counts.Frame, np.zeros(reference['n_residues']))
        assert (counts.Total == reference['n_neighbours']).all()
        assert_array_equal(counts.n0, reference['n_LIPID_neighbours'])
        assert_array_equal(counts.n1, reference['n_CHOL_neighbours'])
        
    def test_count_neighbours_count_by_offset(self, neighbours, reference):
        
        # make every LIPI take the value 0
        # and every CHOL take the value 1
        count_by = np.zeros((reference['n_residues'], reference['n_frames']), dtype=np.int8)
        count_by[1::2] = 1
        
        # check it doesn't matter what numbers are in 'count_by'
        # make LIPI = 100, CHOL = 110
        count_by = (count_by + 10) * 10
        
        counts = neighbours.count_neighbours(count_by=count_by)
        
        assert_array_equal(counts.shape, (reference['n_residues'] * reference['n_frames'], reference['n_columns']))
        assert_array_equal(counts.Frame, np.zeros(reference['n_residues']))
        assert (counts.Total == reference['n_neighbours']).all()
        assert_array_equal(counts.n100, reference['n_LIPID_neighbours'])
        assert_array_equal(counts.n110, reference['n_CHOL_neighbours'])

    def test_count_neighbours_count_by_labels(self, neighbours, reference):
        
        # make every LIPI take the value 0
        # and every CHOL take the value 1
        count_by = np.zeros((reference['n_residues'], reference['n_frames']), dtype=np.int8)
        count_by[1::2] = 1
        
        # liquid-disordered  = 0
        # liquid-ordered = 1
        count_by_labels = {"Ld": 0, "Lo": 1}
        
        counts = neighbours.count_neighbours(count_by=count_by, count_by_labels=count_by_labels)
        
        assert_array_equal(counts.shape, (reference['n_residues'] * reference['n_frames'], reference['n_columns']))
        assert_array_equal(counts.Frame, np.zeros(reference['n_residues']))
        assert (counts.Total == reference['n_neighbours']).all()
        assert_array_equal(counts.nLd, reference['n_LIPID_neighbours'])
        assert_array_equal(counts.nLo, reference['n_CHOL_neighbours'])

    def test_run_method_not_called(self, universe):
        
        neighbours = Neighbours(
            universe=universe,
            lipid_sel='name L C',
            cutoff=12.0
        )
        
        match = ".neighbours attribute is None: use .run\(\) before calling .count_neighbours\(\)"  # noqa:W605
        with pytest.raises(NoDataError, match=match):
            neighbours.count_neighbours()


class TestNeighboursClusters:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)

    kwargs = {
        'lipid_sel': 'name L C',
        'cutoff': 12.0
    }
    
    @pytest.fixture(scope='class')
    def neighbours(self, universe):
        neighbours = Neighbours(universe, **self.kwargs)
        neighbours.run()
        return neighbours
    
    @pytest.fixture(scope='class')
    def reference(self):
        
        reference = {
            'n_frames': 1,
            'largest_leaflet': 50,
            'largest_lipid': 15,
            'largest_chol': 15,
            'all_upper_indices': np.arange(50),
            'all_lower_indices': np.arange(50, 100),
            'lipid_upper_indices': np.array(
                [
                    0, 4, 6, 10, 14, 16, 20, 24, 26, 30,
                    34, 36, 40, 44, 46
                ]
            ),
            'chol_lower_indices': np.array(
                [
                    53, 55, 59, 63, 65, 69, 73, 75, 79, 83,
                    85, 89, 93, 95, 99
                ]
            ),
            'lipid_sel': "name L",
            'chol_sel': 'name C',
            'all_sel': 'name L C',
            'leaflets': np.ones(100, dtype=np.int8)
        }
        
        # first 50 residues in the upper leaflet
        # next 50 in the lower leaflet
        reference['leaflets'][50:] = -1
        
        return reference
    
    def test_clusters(self, neighbours, reference):
        
        largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
            cluster_sel=reference['all_sel'],
            return_resindices=True
        )
        
        assert largest_cluster.size == reference['n_frames']
        assert_array_equal(largest_cluster, reference['largest_leaflet'])
        assert_array_equal(largest_cluster_indices[0], reference['all_upper_indices'])
        
    def tests_clusters_lower_leaflet(self, neighbours, reference):
        
        largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
            cluster_sel=reference['all_sel'],
            filter_by=reference['leaflets'] == -1,
            return_resindices=True
        )
        
        assert largest_cluster.size == reference['n_frames']
        assert_array_equal(largest_cluster, reference['largest_leaflet'])
        assert_array_equal(largest_cluster_indices[0], reference['all_lower_indices'])
        
    def test_clusters_upper_leaflet_lipid(self, neighbours, reference):
        
        largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
            cluster_sel=reference['lipid_sel'],
            filter_by=reference['leaflets'] == 1,
            return_resindices=True
        )
        
        assert largest_cluster.size == reference['n_frames']
        assert_array_equal(largest_cluster, reference['largest_lipid'])
        assert_array_equal(largest_cluster_indices[0], reference['lipid_upper_indices'])
        
    def test_clusters_lower_leaflet_chol(self, neighbours, reference):
        
        largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
            cluster_sel=reference['chol_sel'],
            filter_by=reference['leaflets'] == -1,
            return_resindices=True
        )
        
        assert largest_cluster.size == reference['n_frames']
        assert_array_equal(largest_cluster, reference['largest_chol'])
        assert_array_equal(largest_cluster_indices[0], reference['chol_lower_indices'])
        
    def test_clusters_dont_return_resindices(self, neighbours, reference):
        
        largest_cluster = neighbours.largest_cluster(
            cluster_sel=reference['all_sel'],
            filter_by=reference['leaflets'] == 1,
            return_resindices=False
        )
        
        assert isinstance(largest_cluster, np.ndarray)
        assert largest_cluster.size == reference['n_frames']
        
    def test_no_cluster_sel(self, neighbours, reference):
        
        largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
            filter_by=reference['leaflets'] == 1,
            return_resindices=True
        )
        
        assert largest_cluster.size == reference['n_frames']
        assert_array_equal(largest_cluster, reference['largest_leaflet'])
        assert_array_equal(largest_cluster_indices[0], reference['all_upper_indices'])
        
    def test_no_filter_by(self, neighbours, reference):
        
        largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
            cluster_sel="name L C",
            return_resindices=True
        )
        
        assert largest_cluster.size == reference['n_frames']
        assert_array_equal(largest_cluster, reference['largest_leaflet'])
        assert_array_equal(largest_cluster_indices[0], reference['all_upper_indices'])
        
    def test_filter_by_2D(self, neighbours, reference):
        
        largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
            cluster_sel="name L C",
            filter_by=reference['leaflets'][:, np.newaxis] == 1,
            return_resindices=True
        )
        
        assert largest_cluster.size == reference['n_frames']
        assert_array_equal(largest_cluster, reference['largest_leaflet'])
        assert_array_equal(largest_cluster_indices[0], reference['all_upper_indices'])
        
    def test_run_method_not_called(self, universe):
        
        neighbours = Neighbours(
            universe=universe,
            lipid_sel='name L C',
            cutoff=12.0
        )
        
        match = ".neighbours attribute is None: use .run\(\) before calling .largest_cluster\(\)"  # noqa:W605
        with pytest.raises(NoDataError, match=match):
            neighbours.largest_cluster()

    def test_bad_cluster_sel(self, neighbours, reference):
          
        match = "'cluster_sel' produces atom empty AtomGroup. Please check the selection string."
        with pytest.raises(ValueError, match=match):
            largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
                cluster_sel="",
                filter_by=reference['leaflets'] == -1,
                return_resindices=True
            )
            
    def test_bad_filter_by_dimensions(self, neighbours, reference):
          
        match = "'filter_by' must either be a 1D array containing non-changing boolean"
        with pytest.raises(ValueError, match=match):
            largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
                cluster_sel=reference['all_sel'],
                filter_by=np.array(None),
                return_resindices=True
            )
            
    def test_bad_filter_num_lipids(self, neighbours, reference):
          
        match = "The shape of 'filter_by' must be \(n_residues,\)"  # noqa:W605
        with pytest.raises(ValueError, match=match):
            largest_cluster, largest_cluster_indices = neighbours.largest_cluster(
                cluster_sel=reference['all_sel'],
                filter_by=reference['leaflets'][:99] == 1,
                return_resindices=True
            )




import pytest
import numpy as np
import MDAnalysis

from numpy.testing import assert_array_almost_equal

from lipyphilic._simple_systems.simple_systems import HEX_LAT
from lipyphilic.lib.assign_leaflets import AssignLeaflets
from lipyphilic.lib.registration import Registration
 
 
class TestRegistration:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)
    
    @pytest.fixture(scope='class')
    def leaflets(self, universe):
        leaflets = AssignLeaflets(universe, lipid_sel="name L C")
        leaflets.run()
        return leaflets

    def test_registred(self, leaflets):
    
        registration = Registration(
            leaflets=leaflets,
            upper_sel="name L C",
            lower_sel="name L C"
        )
        registration.run()
    
        reference = {
            'n_frames': 1,
            'registration': 1
        }
    
        assert registration.registration.size == reference['n_frames']
        assert_array_almost_equal(registration.registration, reference['registration'])
        
    def test_antiregistred(self, leaflets):
        
        registration = Registration(
            leaflets=leaflets,
            upper_sel="name L",
            lower_sel="name C"
        )
        registration.run()
    
        reference = {
            'n_frames': 1,
            'registration': -1
        }
    
        assert registration.registration.size == reference['n_frames']
        assert_array_almost_equal(registration.registration, reference['registration'])

    def test_nbins100(self, leaflets):
        
        registration = Registration(
            leaflets=leaflets,
            upper_sel="name L C",
            lower_sel="name L C",
            n_bins=100
        )
        registration.run()
    
        reference = {
            'n_frames': 1,
            'registration': 1
        }
    
        assert registration.registration.size == reference['n_frames']
        assert_array_almost_equal(registration.registration, reference['registration'])
        
    def test_nbins1000_sd0(self, leaflets):
        
        # The bins are 0.1 Anstrom wide, and gaussian_sd=0 means there is no spead
        # in density
        # So, registration should be
        registration = Registration(
            leaflets=leaflets,
            upper_sel="name L",
            lower_sel="name C",
            n_bins=1000,
            gaussian_sd=0
        )
        registration.run()
    
        reference = {
            'n_frames': 1,
            'registration': 0
        }
    
        assert registration.registration.size == reference['n_frames']
        assert_array_almost_equal(registration.registration, reference['registration'], decimal=4)
        
    def test_filter_by_registered(self, leaflets):
        
        filter_by = np.zeros(100)
        filter_by[1::2] = 1
        
        registration = Registration(
            leaflets=leaflets,
            filter_by=filter_by
        )
        registration.run()
    
        reference = {
            'n_frames': 1,
            'registration': 1
        }
    
        assert registration.registration.size == reference['n_frames']
        assert_array_almost_equal(registration.registration, reference['registration'], decimal=4)
        
    def test_filter_by_2D_registered(self, leaflets):
        
        filter_by = np.zeros(100)
        filter_by[1::2] = 1
        
        registration = Registration(
            leaflets=leaflets,
            filter_by=filter_by[:, np.newaxis]
        )
        registration.run()
    
        reference = {
            'n_frames': 1,
            'registration': 1
        }
    
        assert registration.registration.size == reference['n_frames']
        assert_array_almost_equal(registration.registration, reference['registration'], decimal=4)


class TestRegistrationExceptions:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)
    
    @pytest.fixture(scope='class')
    def leaflets(self, universe):
        leaflets = AssignLeaflets(universe, lipid_sel="name L C")
        leaflets.run()
        return leaflets
    
    def test_Exceptions(self, leaflets):
            
        match = "leaflets must be of type AssignLeaflets"
        with pytest.raises(ValueError, match=match):
            Registration(
                leaflets=leaflets.leaflets  # this is a NumPy array
            )
            
        filter_by = np.zeros(100)
        filter_by[1::2] = 1
        match = "'filter_by' must either be a 1D array containing non-changing boolean"
        with pytest.raises(ValueError, match=match):
            Registration(
                leaflets=leaflets,
                filter_by=np.array(None)
            )
            
        match = "The shape of 'filter_by' must be \(n_residues,\)"  # noqa:W605
        with pytest.raises(ValueError, match=match):
            Registration(
                leaflets=leaflets,
                filter_by=filter_by[:99]
            )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# lipyphilic --- lipyphilic.readthedocs.io
#
# Released under the GNU Public Licence, v2 or any higher version
#

"""Assign leaflets --- :mod:`lipyphilic.lib.assign_leaflets`
============================================================

:Author: Paul Smith
:Year: 2021
:Copyright: GNU Public License v2

This module provides methods for assigning lipids to leaflets in a bilayer.

The class :class:`lipyphilic.lib.assign_leaflets.AssignLeaflets` assigns
each lipid to a leaflet based on the distance in *z* to the midpoint of
the bilayer. Lipids may be assigned to the upper leaflet (indicated by `1`),
the lower leaflet (`-1`) or the bilayer midplane (`0`).

Input
------

Required:
  - *universe* : an MDAnalysis Universe object
  - *lipid_sel* : atom selection for *all* lipids in the bilayer, including e.g. sterols

Options:
  - *midplane_sel* : atom selection for lipid that may occupy the midplane
  - *midplane_cutoff* : atoms within this distance from the midpoint are considered to be the midplane
  - *n_bins* : split the membrane into *n_bins \\* n_bins* patches, and calculate local membrane midpoints for each patch

Output
------

  - *leaflets* : leaflet to which each lipid is assigned at each frame
  
Leaflet data are returned in a :class:`numpy.ndarray`, where each row corresponds
to an individual lipid and each column corresponds to an individual frame, i.e.
leaflets[i, j] refers to the leaflet of lipid *i* at frame *j*. The results are
accessible via the `AssignLeaflets.leaflets` attribute.


Example usage of :class:`AssignLeaflets`
----------------------------------------

An MDAnalysis Universe must first be created before using AssignLeaflets::

  import MDAnalysis as mda
  from lipyphilic.lib.assign_leaflets import AssignLeaflets

  u = mda.Universe(tpr, trajectory)

If we have used the MARTINI forcefield to study phospholipid/cholesterol mixture,
we can assign lipids and cholesterol to the upper and lower as follows::

  leaflets = AssignLeaflets(
    universe=u,
    lipid_sel="name GL1 GL2 ROH"
  )
  
We then select which frames of the trajectory to analyse (`None` will use every
frame) and choose to display a progress bar (`verbose=True`)::
  
  leaflets.run(
    start=None,
    stop=None,
    step=None,
    verbose=True
  )
  
The results are then available in the :attr:`leaflets.leaflets` attribute as a
:class:`numpy.ndarray`. Each row corresponds to an individual lipid and each column
to an individual frame, i.e `leaflets.leaflets[i, j]` contains the leaflet
membership of lipid *i* at frame *j*. Lipid *i*, at frame *j*, is in the upper
leaflet if `leaflets.leaflets[i, j]==1` and in the lower leaflet if
`leaflets.leaflets[i, j]==-1`.

The above example will assign every lipid (including sterols) to either the upper
or lower leaflet. To allow cholesterol to be in the midplane, we can provide
a :attr:`midplane_sel` and :attr:`midplane_cutoff` to :class:`AssignLeaflets`::

  leaflets = AssignLeaflets(
    universe=u,
    lipid_sel="name GL1 GL2 ROH",
    midplane_sel="resname CHOL and name ROH C2",
    midplane_cutoff=12.0
  )
  
A cholesterol molecule that has both its *ROH* and *C2* atoms within *12* Å of
membrane midpoint will be assigned to the midplane, i.e. for cholesterol *i*
at frame *j* that is in the midplane, `leaflets.leaflets[i, j]==0`.

The first two examples compute a global membrane midpoint based on all the atoms
of the lipids in the. Lipids are then assigned a leaflet based on their distance
in *z* to this midpoint. This is okay for planar bilayers, but can lead to incorrect
leaflet classification in membranes with large undulations. If your bilayer has
large undulations, `AssignLeaflets` account for this by creating a grid in *xy*
patches of your membrane, calculating the local membrane midpoint in each patch,
the assign leaflet membership based on distance in *z* to the local membrane
midpoint. This is done through use of `n_bins`::

  leaflets = AssignLeaflets(
    universe=u,
    lipid_sel="name GL1 GL2 ROH",
    midplane_sel="resname CHOL and name ROH C2",
    midplane_cutoff=12.0,
    n_bins=10
  )
  
In this example, the membrane will be split into a *10 x 10* grid and a lipid
assigned a leaflet based on the distance to the midpoint of the patch the lipid
is in.

The class and its methods
-------------------------

.. autoclass:: AssignLeaflets
    :members:

"""

import numpy as np
import scipy.stats

from lipyphilic.lib import base


class AssignLeaflets(base.AnalysisBase):
    """Assign lipids in a bilayer to the upper leaflet, lower leaflet, or midplane.
    """

    def __init__(self, universe,
                 lipid_sel,
                 midplane_sel=None, midplane_cutoff=None,
                 n_bins=1):
        """Set up parameters for assigning lipids to a leaflet.

        Parameters
        ----------
        universe : Universe
            MDAnalysis Universe object
        lipid_sel : str
            Selection string for the lipids in a membrane. The selection
            should cover **all** residues in the membrane, including cholesterol.
        midplane_sel :  str, optional
            Selection string for residues that may be midplane. Any residues not
            in this selection will be assigned to a leaflet regardless of its
            proximity to the midplane.
            The default is `None`, in which case all lipids will be assigned to
            either the upper or lower leaflet.
        midplane_cutoff : float, optional
            Minimum distance in *z* an atom must be from the midplane to be assigned
            to a leaflet rather than the midplane. The default is `0`, in which case
            all lipids will be assigned to either the upper or lower leaflet. Must
            be non-negative.
        n_bins : int, optional
            Number of bins in *x* and *y* to use to create a grid of membrane patches.
            Local membrane midpoints are computed for each patch, and lipids assigned
            a leaflet based on the distance to their local membrane midpoint. The
            default is `1`, which is equivalent to computing a single global
            midpoint.
            
        Note
        ----

        Typically, :attr:`midplane_sel` should select only sterols. Other lipids have
        flip-flop rates that are currently unaccessible with MD simulations, and thus
        should always occupy either the upper or lower leaflet.
        """
        super(AssignLeaflets, self).__init__(universe.trajectory)

        self.u = universe
        self.membrane = self.u.select_atoms(lipid_sel, updating=False)

        if (midplane_sel is not None) ^ (midplane_cutoff is not None):
            raise ValueError(f"midplane_sel is '{midplane_sel}' and midplane_cutoff "
                             f"is {midplane_cutoff}. To assign molecules to the midplane, "
                             "midplane_sel must be provided and midplane_cutoff must be "
                             "greater than 0."
                             )
        
        if (midplane_cutoff is not None) and (midplane_cutoff <= 0):
            raise ValueError("To assign molecules to the midplane, midplane_cutoff must"
                             "be greater than 0."
                             )
        
        self.potential_midplane = self.u.select_atoms(midplane_sel, updating=False) if midplane_sel else None
        self.midplane_cutoff = midplane_cutoff if midplane_cutoff else 0.0
        
        if self.potential_midplane and ((self.potential_midplane - self.membrane.residues.atoms).n_atoms > 0):
            raise ValueError("midplane_sel contains atoms that are not present in molecules selected "
                             "in lipid_sel. lipid_sel must cover *all* residues in the membrane."
                             )

        self.n_bins = n_bins
        self.leaflets = None
        
    def _prepare(self):
        
        # Output array
        self.leaflets = np.full(
            (self.membrane.n_residues, self.n_frames),
            fill_value=0,
            dtype=np.int8  # smallest sized `np.int` is 1 byte, still 8 times smaller than using `int`
        )

    def _single_frame(self):
        
        # Atoms must be wrapped before creating a lateral grid of the membrane
        self.membrane.wrap(inplace=True)

        # Find the midpoint of the bilayer as a function of (x,y), using
        # `n_bins` grid points in each dimensions
        # Use all atoms in the membrane to get better statistics
        if self.n_bins > 1:
            bins = np.linspace(0.0, self._ts.dimensions[0], self.n_bins + 1)
        else:
            # scipy.stats.binned_statistics raises Value error if there is only one bin
            bins = [0.0, self._ts.dimensions[0] + 1, self._ts.dimensions[0] + 2]
        
        memb_midpoint_xy = scipy.stats.binned_statistic_2d(
            x=self.membrane.residues.atoms.positions[:, 0],
            y=self.membrane.residues.atoms.positions[:, 1],
            values=self.membrane.residues.atoms.positions[:, 2],
            statistic="mean",
            bins=bins,
            expand_binnumbers=True
        )
        
        # Assign leaflets
        self._assign_leaflets(memb_midpoint_xy)
        
        # if necessary, find midplane residues
        if (self.potential_midplane is not None) and self.midplane_cutoff > 0.0:
            self._find_midplane(memb_midpoint_xy=memb_midpoint_xy)
    
    def _assign_leaflets(self, memb_midpoint_xy):
        """Assign lipids to the upper (1) or lower (-1) leaflet.

        Parameters
        ----------
        memb_midpoint_xy : BinnedStatistic2dResult
            Membrane grid created with stats.binned_statistic_2d. Contains the midpoint of
            each membrane patch.
        """
        
        # x and y have the same number of bins
        bins = memb_midpoint_xy.x_edge
        
        # get the binnumbers for each lipid
        lipid_x_bins, lipid_y_bins = scipy.stats.binned_statistic_2d(
            x=self.membrane.positions[:, 0],
            y=self.membrane.positions[:, 1],
            values=self.membrane.positions[:, 2],
            statistic="mean",
            bins=bins,
            expand_binnumbers=True
        ).binnumber -1  # These were bin numbers, now bin indices  # noqa: E225
        
        upper_leaflet = self.membrane[
            self.membrane.positions[:, 2] >
            (memb_midpoint_xy.statistic[lipid_x_bins, lipid_y_bins])  # we don't to consider midplane_cutoff here
        ]
        self.leaflets[
            np.in1d(self.membrane.residues.resindices, upper_leaflet.residues.resindices),
            self._frame_index
        ] = 1
        
        lower_leaflet = self.membrane[
            self.membrane.positions[:, 2] <
            (memb_midpoint_xy.statistic[lipid_x_bins, lipid_y_bins])  # we don't to consider midplane_cutoff here
        ]
        self.leaflets[
            np.in1d(self.membrane.residues.resindices, lower_leaflet.residues.resindices),
            self._frame_index
        ] = -1
        
        return None
         
    def _find_midplane(self, memb_midpoint_xy):
        """Determine which residues are in the midplane

        Parameters
        ----------
        memb_midpoint_xy : BinnedStatistic2dResult
            Membrane grid created with stats.binned_statistic_2d. Contains the midpoint of
            each membrane patch.
        """
        
        # Atoms must be wrapped before so we can assign lipids to grid patches
        self.potential_midplane.wrap(inplace=True)
        
        # x and y have the same number of bins
        bins = memb_midpoint_xy.x_edge
        
        midplane_x_bins, midplane_y_bins = scipy.stats.binned_statistic_2d(
            x=self.potential_midplane.positions[:, 0],
            y=self.potential_midplane.positions[:, 1],
            values=self.potential_midplane.positions[:, 2],
            statistic="mean",
            bins=bins,
            expand_binnumbers=True
        ).binnumber -1  # These were bin numbers, now bin indices  # noqa: E225
        
        # First assume they're all midplane
        # Then find residues that have at least one atom further than
        # `midplane_cutoff` from the local midplane
        midplane_mask = np.full(self.potential_midplane.n_residues, fill_value=True, dtype=bool)

        not_midplane = np.abs(
            self.potential_midplane.positions[:, 2] - memb_midpoint_xy.statistic[midplane_x_bins, midplane_y_bins]
        ) > self.midplane_cutoff

        # These residues have at least one atom in `potential_midplane`
        # that is more the `midplane_cutoff` from the local midplane
        midplane_mask[
            np.in1d(self.potential_midplane.residues.resindices, self.potential_midplane[not_midplane].resindices),
        ] = False

        midplane_residues = self.potential_midplane.residues[midplane_mask]

        # Assign midplane
        self.leaflets[
            np.in1d(self.membrane.residues.resindices, midplane_residues.resindices),
            self._frame_index
        ] = 0
        
        return None



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# lipyphilic --- lipyphilic.readthedocs.io
#
# Released under the GNU Public Licence, v2 or any higher version
#

"""Neighbours --- :mod:`lipyphilic.lib.neighbours`
==================================================

:Author: Paul Smith
:Year: 2021
:Copyright: GNU Public License v2

This module provides methods for finding neighbouring lipids in a bilayer,
calculating local lipid compositions and finding the largest cluster of
lipids over time.

Two lipids are considered neighbours if they have any atoms within a given
cutoff distance of one another.

Input
-----

Required:
  - *universe* : an MDAnalysis Universe object.
  - *lipid_sel* : atom selection for lipids in the bilayer

Optional:
  - *cutoff* : lipids are considered to be neighbouring if they have at least one pair of atoms less than this distance apart (in Å)
  

Output
------

  - *neighbouring* : a binary variable, equal to 1 if two lipids are in contact, and 0 otherwise

For efficient use of memory, an adjacency matrix of neighbouring lipids is stored
in a :class:`scipy.sparse.csc_matrix` sparse matrix for each frame of the analysis. The data
are stored in the :attr:`neighbours.neighbours` attribute. The matrix has shape
(n_residues, n_residues * n_frames), and the data for frame *n* is accessed via::

  neighbours.neighbours[:, (n * n_residues):((n + 1) * n_residues)]

A matrix element *[i, j]* is equal to 1 if lipid *i* neighbours lipid *j* and equal to 0 otherwise.
The matrix is symmetric: if lipid *i* neighbours lipid *j* then *j* must neighbour *i*.


Tip
---

The resultant sparse matrix can be used to calculate the number of each lipid species
around each individual lipid at each frame using :func:`lipyphilic.lib.neighbours.count_neighbours`,
or to find the largest cluster of lipids at each frame using
:func:`lipyphilic.lib.neighbours.largest_cluster`.


Example usage of :class:`Neighbours`
------------------------------------

An MDAnalysis Universe must first be created before using :class:`Neighbours`::

  import MDAnalysis as mda
  from lipyphilic.lib.neighbours import Neighbours

  u = mda.Universe(tpr, trajectory)

We can now create our :class:`Neighbours` object::

  neighbours = Neighbours(
      universe=u,
      lipid_sel="name GL1 GL2 ROH",  # assuming we're using the MARTINI forcefield
      cutoff=12.0
  )
  
A lipid will be considered to be neighbouring a cholesterol molecule if either its *GL1* or *GL2* bead
is within *12* Å of the ROH bead of the cholesterol. For neighbouring lipids, the distances
between there respective *GL1* and "GL2* beads will be considered.
  
We then select which frames of the trajectory to analyse (`None` will use every
frame) and select to display a progress bar (`verbose=True`)::
  
  neighbours.run(
    start=None,
    stop=None,
    step=None,
    verbose=True
  )
  
The results are then available in the :attr:`neighbours.Neighbours` attribute as a
:class:`scipy.sparse.csc_matrix`.

Counting the number of neighbours: by lipid species
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

In order to compute the number of each lipid species around each lipid at each frame,
after generating the neighbour matrix we can use the :func:`Neighbours.count_neighbours`
method::

  counts = neighbours.count_neighbours()

Counts is a :class:`pandas.DataFrame` in which each row contains the following
information (if there are N distinct species in the membrane)::

    [
        <lipid identifier>,  # by default, the lipid resname
        <lipid resindex>,
        <frame>,
        <num species_1 neighbours>,
        ...
        <num species_N neighbours>,
        <total num neighbours>
    ]

Counting the number of neighbours: by user-defined labels
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Instead of using the lipid resname to identify neighbouring lipids, any ordinal data may
be used for counting lipid neighbours through use of the :attr:`count_by` and
:attr:`count_by_labels` parameters::

  counts = neighbours.count_neighbours(
    count_by=lipid_order_data,
    count_by_labels={'Ld': 0, 'Lo': 1}
  )

Here we assume that 'lipid_order_data' contains information on whether each lipid is in
the liquid-disordered phase or the liquid-ordered phase at each frame. It must take
the shape '(n_residues, n_frames)', and in this example 'lipid_order_data[i, j]' would
be equal to zero if lipid 'i' is liquid-disordered at frame 'j' and equal to 1 if it is
liquid-ordered. 'count_by_labels' is used to signify that the value '0' corresponds to
the liquid-disordered (Ld) phase and the value '1' to the liquid-ordered  (Lo) phase. In
this example, the returned :class:`pandas.DataFrame` would contain the following information
in each row::
  
    [
        <Ld or Lo>,
        <lipid resindex>,
        <frame>,
        <num Ld neighbours>,
        <num Lo neighbours>,
        <total num neighbours>
    ]


Find the largest cluster
^^^^^^^^^^^^^^^^^^^^^^^^

To find the largest cluster of a set of lipid species we can use the :func:`Neighbours.largest_cluster`
method::

  largest_cluster = neighbours.largest_cluster(
      cluster_sel="resname CHOL DPPC"
  )
  
The results are returned in a :class:`numpy.ndarray` and contain the number of lipids in the largest
cluster at each frame.

  
Find the largest cluster in a given leaflet
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The previous example will compute the largest cluster formed by cholesterol and DPPC molecules at each
frame. In large coarse-grained systems where there is substantial flip-flop of sterols, this cluster may
span both leaflets. In order to find the largest cluster at each frame within a given leaflet, we can
tell :func:`Neighbours.largest_cluster` to consider only lipids in the upper leaflet by using the
`filter_by` parameter.

First, though, we need to know which leaflet each lipid is in at each frame. This may be done using
:class:`lipyphilic.lib.assign_leaflets.AssignLeaflets`::

  leaflets = AssignLeaflets(
    universe=u,
    lipid_sel="name GL1 GL2 ROH"  # pass the same selection that was passed to Neighbours
  )
  leaflets.run()  # run the analysis on the same frames as Neighbours.run()
  
The leaflets data are stored in the :attr:`leaflets.leaflets` attribute, will be equal to '1' if the
lipid is in the upper leaflet at a given frame and equal to '-1' if it is in the lower leaflet. See
:class:`lipyphilic.lib.assign_leaflets.AssignLeaflets` for more information. We can now find the
largest cluster over time in the upper (1) leaflet.

The :attr:`filter_by` parameter takes as input a 2D :class:`numpy.ndarray` of shape
(n_residues, n_frames). The array should be a `boolean mask
<https://docs.scipy.org/doc/numpy-1.15.0/user/basics.indexing.html#boolean-or-mask-index-arrays>`__,
where `True` indicates that we should include this lipid in the neighbour calculation::

  upper_leaflet_mask = leaflet.leaflets == 1

  largest_cluster_upper_leaflet = neighbours.largest_cluster(
      cluster_sel="resname CHOL DPPC",
      leaflets=upper_leaflet_mask
  )

Now, lipids either in the lower leaflet (-1) or the midplane (0) will not be included when determining
the largest cluster.

Get residue indices of lipids in the largest cluster
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

It we want to know not just the cluster size but also which lipid are in the largest cluster at each
frame, we can set the :attr:`return_resindices` parameter to `True`::

  largest_cluster, largest_cluster_resindices = neighbours.largest_cluster(
      cluster_sel="resname CHOL DPPC",
      return_resindices=True
  )

The resindices will be returned as list of `numpy.ndarray` arrays - one per frame of the analysis. Each
array contains the resindices of the lipids in the largest cluster at that frame


The class and its methods
-------------------------

.. autoclass:: Neighbours
    :members:

"""
from tqdm.auto import tqdm
import numpy as np
import scipy.stats
import scipy.sparse
import pandas as pd
from MDAnalysis.lib.distances import capped_distance
from MDAnalysis.exceptions import NoDataError

from lipyphilic.lib import base


class Neighbours(base.AnalysisBase):
    """Find neighbouring lipids in a bilayer.
    """

    def __init__(self, universe,
                 lipid_sel,
                 cutoff=10.0
                 ):
        """Set up parameters for finding neighbouring lipids.
        
        Parameters
        ----------
        universe : Universe
            MDAnalysis Universe object
        lipid_sel : str
            Selection string for lipids in the bilayer.
        cutoff : float, optional
            To be considered neighbours, two lipids must have at least one pair of atoms within
            this cutoff distance (in Å). The default is `10.0`.
        
        """
        super(Neighbours, self).__init__(universe.trajectory)
        
        self.u = universe
        self.membrane = self.u.select_atoms(lipid_sel, updating=False)
        
        # to allow for non-sequential resindices
        self._sorted_membrane_resindices = scipy.stats.rankdata(
            self.membrane.resindices,
            method="dense"
        ) - 1
        
        if cutoff <= 0:
            raise ValueError("'cutoff' must be greater than 0")
        
        self.cutoff = cutoff
           
        self.neighbours = None
        
    def _prepare(self):
        
        # Output array
        self.neighbours = scipy.sparse.lil_matrix(
            (self.membrane.n_residues, self.membrane.n_residues * self.n_frames),
            dtype=np.int8
        )
        
    def _single_frame(self):
        
        pairs = capped_distance(
            self.membrane.positions,
            self.membrane.positions,
            max_cutoff=self.cutoff,
            box=self._ts.dimensions,
            return_distances=False
        )
        self.pairs = pairs
        
        # Find unique pairs of residues interacting
        # Currently we have pairs of atoms
        ref, neigh = np.unique(self._sorted_membrane_resindices[pairs], axis=0).T
        
        # Dont keep self-interactions between lipids
        different = ref != neigh
        ref = ref[different]
        neigh = neigh[different]
        
        # store neighbours for this frame
        frame_start = self._frame_index * self.membrane.n_residues
        self.neighbours[ref, neigh + frame_start] = 1
        self.neighbours[neigh, ref + frame_start] = 1  # the neighbour matrix must be symmetric

    def _conclude(self):
        
        # csc sparse matrices are faster for slicing columns
        # we'll want this for selecting neighbours at given frames
        self.neighbours = self.neighbours.tocsc()

    def count_neighbours(self, count_by=None, count_by_labels=None):
        """Count the number of each neighbour type at each frame.

        Parameters
        ----------
        count_by : numpy.ndarray, optional
            An array containing ordinal data describing each lipid at each frame. For example,
            it may be an array containing information on the ordered state or each lipid.
            Defaults to None, in which case the lipid species (resnames) are used for counting neighbours.
        count_by_labels : dict, optional
            A dictionary of labels describing what each unique value in `count_by` refers to, e.g
            if `count_by` contains information on the ordered state of each lipid at each frame, whereby
            0 corresponds to disordered and 1 corresponds to ordered, then
            `count_by_labels = {'Ld': 0, 'Lo': 1}`. There **must** be precisely one label for each unique
            value in 'count_by'. If `count_by` is given but `count_by_labels` is left as `None`, the values
            in `count_by` will be used as the labels.
        
        Returns
        -------
        
        counts : pandas.DataFrame
            A DataFrame containing the following data for each lipid at each frame: lipid identifier
            (default is resname), lipid resindex, frame number, number of neighbours of each species
            (or of each type in 'count_by' if this is provided), as well as the total number of neighbours.
        
        """
        
        if self.neighbours is None:
            raise NoDataError(".neighbours attribute is None: use .run() before calling .count_neighbours()")
        
        # create output array
        if count_by is None:
            
            # Use lipid resnames to distinguish lipids
            count_by = np.full(
                (self.membrane.n_residues, self.n_frames),
                fill_value=self.membrane.residues.resnames[:, np.newaxis],
            )
            count_by_labels = {label: index for index, label in enumerate(np.unique(self.membrane.resnames))}
        
        elif count_by_labels is None:
            
            # Use values in 'count_by' as the labels
            count_by_labels = {label: index for index, label in enumerate(np.unique(count_by))}
            
        else:
            
            # the ordinal values in 'count_by' now take on the string labels supplied
            max_label_size = max([len(label) for label in count_by_labels])
            new_count_by = np.full_like(count_by, dtype=f'<U{max_label_size}', fill_value="")
            for label in count_by_labels:
                new_count_by[count_by == count_by_labels[label]] = label
            count_by = new_count_by
            del new_count_by
        
        # create output array
        all_counts = np.full(
            (self.membrane.n_residues, self.n_frames, len(count_by_labels)),
            fill_value=0,
            dtype=np.uint8  # count can't be negative, and no lipid will have more than 255 neighbours
        )
        
        # For counts we need to know which column of the output array to add counts to for each lipid type
        type_index = {value: index for index, value in enumerate(count_by_labels)}
        
        # Get counts at each frame
        n_residues = self.membrane.n_residues
        for frame_index in tqdm(np.arange(self.n_frames)):
        
            ref, neigh = self.neighbours[:, frame_index * n_residues:(frame_index + 1) * n_residues].nonzero()
            unique, counts = np.unique([ref, [type_index[t] for t in count_by[neigh, frame_index]]], axis=1, return_counts=True)
            
            r, t = unique  # reference index (r) and type index (t)
            all_counts[r, frame_index, t] = counts

        # Assemble data for the DataFrame
        labels = np.array([list(count_by_labels)[type_index[frame_index]] for lipid in count_by for frame_index in lipid])
        
        resindices = np.full((n_residues, self.n_frames), fill_value=self.membrane.residues.resindices[:, np.newaxis])
        resindices = resindices.reshape(n_residues * self.n_frames)
        
        frames = np.full((n_residues, self.n_frames), fill_value=self.frames)
        frames = frames.reshape(n_residues * self.n_frames)

        all_counts = all_counts.reshape(n_residues * self.n_frames, len(count_by_labels))
        total_counts = np.sum(all_counts, axis=1)
        
        data = np.concatenate(
            (labels[:, np.newaxis], resindices[:, np.newaxis], frames[:, np.newaxis], all_counts, total_counts[:, np.newaxis]),
            axis=1
            )
        
        # Create DataFrame
        columns = ["Label", "Resindex", "Frame"] + [f"n{label}" for label in count_by_labels] + ["Total"]
        df = pd.DataFrame(
            data=data,
            columns=columns
        )
        
        # make every column except the label take on integer values
        for column in df.columns[1:]:
            df[column] = pd.to_numeric(df[column])
        
        return df
    
    def largest_cluster(self, cluster_sel=None, filter_by=None, return_resindices=False):
        """Find the largest cluster of lipids at each frame.
        
        Parameters
        ----------
        cluster_sel : str, optional
            Selection string for lipids to include in the cluster analysis. The default is `None`, in
            which case all lipid used in identiying neighbouring lipids will be used for finding
            the largest cluster.
        filter_by : numpy.ndarray, optional
            A boolean array indicating whether or not to include each lipid in the cluster analysis. If
            the array is 1D and of shape (n_lipids), the same lipids will be used in the cluster
            analysis at every frame. If the array is 2D and of shape (n_lipids, n_frames), the boolean
            value of each lipid at each frame will be taken into account. The default is `None`, in which
            case all lipids used in identiying neighbours will be used for finding
            the largest cluster.
        return_resindices : bool, optional
            If `True`, a list of NumPy arrays will also be returned, on for each frame. Each NumPy array
            will contain the residue indices of the lipids in the largest cluster at that frame. Note, if
            there are two largest clusters of equal size, only the resindices of lipids in one
            cluster will be returned (the cluster that has the lipid with the smallest residue index). The
            default is `False`, in which case no reidue indices are returned.
            
        
        Returns
        -------
        
        largest_cluster : numpy.ndarray
            An array containing the number of lipids in the largest cluster at each frame.
            
        Note
        ----
        
        Neighbours must be found by using `Neighbours.run()` before calling either
        `Neighbours.count_neighbours()` or `Neighbours.largest_cluster()`.
        
        """

        if self.neighbours is None:
            raise NoDataError(".neighbours attribute is None: use .run() before calling .largest_cluster()")
        
        if filter_by is not None and np.array(filter_by).ndim not in [1, 2]:
            raise ValueError("'filter_by' must either be a 1D array containing non-changing boolean"
                             "values for each lipid, or a 2D array of shape (n_residues, n_frames)"
                             " containing a boolean value for each lipid at each frame."
                             )

        elif filter_by is not None and len(filter_by) != self.membrane.n_residues:
            raise ValueError("The shape of 'filter_by' must be (n_residues,)")
        
        # determine which lipids to use in the analysis at each frame
        if filter_by is None:
            
            filter_by = np.full(
                (self.membrane.n_residues, self.n_frames),
                fill_value=True,
                dtype=bool
            )
        elif filter_by.ndim == 1:
            
            filter_by = np.full(
                (self.membrane.n_residues, self.n_frames),
                fill_value=filter_by[:, np.newaxis],
                dtype=bool
            )
            
        # also create mask based on `cluster_sel`
        if cluster_sel is None:
            
            filter_lipids = np.full(
                self.membrane.n_residues,
                fill_value=True,
                dtype=bool
            )
        else:
            
            lipids = self.u.select_atoms(cluster_sel).residues
            
            if lipids.n_residues == 0:
                raise ValueError(
                    "'cluster_sel' produces atom empty AtomGroup. Please check the selection string."
                )
            
            filter_lipids = np.in1d(
                self.membrane.residues.resindices,
                lipids.resindices
            )
            
        # combine the masks
        filter_by[filter_lipids == False] = False  # noqa: E712
                
        # output arrays
        largest_cluster = np.zeros(self.n_frames, dtype=int)
        largest_cluster_resindices = np.full(self.n_frames, fill_value=0, dtype=object)

        n_residues = self.membrane.n_residues
        for frame_index, frame in enumerate(self.frames):
            
            frame_filter = filter_by[:, frame_index]
            
            frame_neighbours = self.neighbours[:, frame_index * n_residues:(frame_index + 1) * n_residues]
            frame_neighbours = frame_neighbours[frame_filter][:, frame_filter].toarray()
            
            # find all connected components
            _, com_labels = scipy.sparse.csgraph.connected_components(frame_neighbours)
            
            unique_com_labels, counts = np.unique(com_labels, return_counts=True)
            largest_label = unique_com_labels[np.argmax(counts)]
            
            # largest cluster and resindices of lipids in the cluster
            largest_cluster[frame_index] = max(counts)
            
            frame_resindices = self.membrane.residues.resindices[frame_filter]
            largest_cluster_resindices[frame_index] = frame_resindices[com_labels == largest_label]
            
        if return_resindices is True:
            return largest_cluster, largest_cluster_resindices
        else:
            return largest_cluster



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# lipyphilic --- lipyphilic.readthedocs.io
#
# Released under the GNU Public Licence, v2 or any higher version
#

r"""Registration --- :mod:`lipyphilic.lib.registration`
======================================================

:Author: Paul Smith
:Year: 2021
:Copyright: GNU Public License v2

This module provides methods for determining registration of leaflets in a bilayer.

The degree of registration is calculated as the pearson correlation coefficient of
densities in the upper and lower leaflets. First, the 2D density of each leaflet,
:math:`L`, is calculated:

.. math::

  \rho(x, y)_{L} = \displaystyle \int\limits_{-\infty}^{\infty} \frac{1}{\sqrt{2\pi}\sigma} \exp \Bigg({-}\frac{1}{2} \bigg(\frac{x' - x}{\sigma} \bigg)^2 \Bigg) \,dx' dy'


where the :math:`(x, y)` positions of lipid atoms in leaflet :math:`L` are binned
into two-dimensional histograms, then convolved with a circular Gaussian density
of standard deviation :math:`\sigma`. :math:`L` is either the upper (:math:`u`) or
lower (:math:`l`) leaflet.

The correlation between the two leaflets, :math:`r_{u/l}`, is then calculated as
the pearson correlation coefficient between :math:`\rho(x, y)_{u}` and
:math:`\rho(x, y)_{l}`. For more information on interleaflet registration in bilayers
see `Thallmair et al. (2018) <https://pubs.acs.org/doi/abs/10.1021/acs.jpclett.8b01877>`__.

Input
-----

Required:
  - *leaflets* : a :class:`lipyphilic.lib.assign_leaflets.AssignLeaflets` object

Optional:
  - *upper_sel* : atom selection for lipids in the upper leaflet to use in the registration calculation
  - *lower_sel* : atom selection for lipids in the lower leaflet to use in the registration calculation
  - *filter_by* : boolean mask for determining which lipids to include in the registration calculation
  - *n_bins* : the number of bins to use in *x* and *y* for the 2D histogram
  - *gaussian_sd* : the standard deviation of the circular Gaussian to convole with the grid densities
  
Output
------

  - *registration* : the degree of interleaglet registration at each frame

The data are stored in the :attr:`registration.registration` attribute, containing the pearson
correlation coefficient of the two-dimensional leaflet densities at each frame.

Example usage of :class:`Registration`
--------------------------------------

An MDAnalysis Universe must first be created before using :class:`Registration`::

  import MDAnalysis as mda
  from lipyphilic.lib.neighbours import Neighbours

  u = mda.Universe(tpr, trajectory)


Then we need to know which leaflet each lipid is in at each frame. This may be done using
:class:`lipyphilic.lib.assign_leaflets.AssignLeaflets`::

  leaflets = AssignLeaflets(
    universe=u,
    lipid_sel="name GL1 GL2 ROH" # assuming we are using the MARTINI forcefield
  )
  leaflets.run()


The leaflets data are stored in the :attr:`leaflets.leaflets` attribute. We can now create our
Registration object by passing our :class:`lipyphilic.lib.assign_leaflets.AssignLeaflets`
object to :class:`Registration` along with atom selections for the lipids::

  registration = Registration(
      leaflets=leaflets,
      upper_sel="resname CHOL and name ROH",
      lower_sel="resname CHOL and name ROH",
  )
  
To calculate the interleaflet correlation of cholesterol molecules using their ROH
beads we then need to use the :func:`run()` method::
  
  registration.run(
    verbose=True
  )

The results are then available in the :attr:`registration.registration` attribute as a
:class:`numpy.ndarray`.

Note
----

The same MDAndalysis Universe and the same frames of the trajectory as used in assigning
leaflets will be used for calculating the interleaflet registration.


Selecting a subset of lipids for the registration analysis
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The previous example will compute the largest registration of cholesterol across the upper
and lower leaflets. In, for example, simulations of phase-separation domains, it is useful
to know the registration of liquid-ordered domains (regardless of the species in the domain)
rather than the registrtion of specific lipid species.

If we have a 2D array, 'lipid_order_data', that contains information on whether each lipid is in
the liquid-disordered phase or the liquid-ordered phase at each frame, we can used this to
calculate the registration of ordered domains. The array must take the shape
'(n_residues, n_frames)', and in the below example 'lipid_order_data[i, j]' would be equal to zero
if lipid 'i' is liquid-disordered at frame 'j' and equal to 1 if it is liquid-ordered::

  registration = Registration(
      leaflets=leaflets,
      filter_by=lipid_order_data == 1
  )


If we have a ternary mixture of DPPC/DOPC/Cholesterol, we can also specifc that we wish to
consider only DPPC and cholesterol in the liquid-ordered phase::

  registration = Registration(
      leaflets=leaflets,
      upper_sel="(resname CHOL and name ROH) or (resname DPPC and name PO4)",
      lower_sel="(resname CHOL and name ROH) or (resname DPPC and name PO4)",
      filter_by=lipid_order_data == 1
  )
  

Changing the resolution of the 2D grid
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

By default, the lipid positions of each leaflet are binned into a two-dimensional
histogram using :math:`n\_bins_x = \lceil{x}\rceil`, where :math:`n\_bins_x` is the
numer of bins in *x* and \lceil{x}\rceil is the size of system in *x* rounded to the
nearest integer. This gives a grid resolution of  *1* Å.

It is also possible to specify the number of bins to use for binning the data::

  registration = Registration(
      leaflets=leaflets,
      upper_sel="resname CHOL and name ROH",
      lower_sel="resname CHOL and name ROH",
      n_bins=100
  )


This will use *100* bins for creating the two-dimensional histogram. Fewer bins
will result in a performance increase but at the cost of spatial resolution. For
all but the largest systems, the default of *1* Å is appropriate. If your system
is larger than a few hundred nm in one dimension, you will likely want to set
:attr:`n_bins` to 2000 or less.

Changing the standard deviation of the circular Gaussian density
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The defualt value of :math`\sigma` is *15*, which is the value used by
`Thallmair et al. (2018) <https://pubs.acs.org/doi/abs/10.1021/acs.jpclett.8b01877>`__
for determining interleaflet cholesterol correlations. This deault value can be
changed using the :attr:`gaussian_sd` parameter::

  registration = Registration(
      leaflets=leaflets,
      upper_sel="resname CHOL and name ROH",
      lower_sel="resname CHOL and name ROH",
      gaussian_sd=12
  )


The obtained correlations are relatively insensitive to changes in :attr:`gaussian_sd`,
and it is recommended to use the deault of *15*.

The class and its methods
-------------------------

.. autoclass:: Registration
    :members:

"""
from tqdm.auto import tqdm
import numpy as np
import scipy

from lipyphilic.lib import base
from lipyphilic.lib.assign_leaflets import AssignLeaflets


class Registration(base.AnalysisBase):
    """Calculate interleaflet registration in a bilayer.
    """

    def __init__(self,
                 leaflets,
                 upper_sel=None,
                 lower_sel=None,
                 filter_by=None,
                 n_bins=None,
                 gaussian_sd=15
                 ):
        """Set up parameters for the registration calculation.
        
        Parameters
        ----------
        leaflets : AssignLeaflets
            An AssignLeaflets object that contains the leaflet membership for each lipid
            at each frame. See lipyphilic.lib.assign_leaflets.AssignLeaflets for details.
        upper_sel : str, optional
            Selection string for lipids in the upper leaflet of the bilayer to be used
            for determining registration. The default is `None`, in which case all atoms
            of all lipids in the upper leaflet will be used.
        lower_sel : str, optional
            Selection string for lipids in the lower leaflet of the bilayer to be used
            for determining registration. The default is `None`, in which case all atoms
            of all lipids in the lower leaflet will be used.
        filter_by : numpy.ndarray, optional
            A boolean array indicating whether or not to include each lipid in the registration
            analysis. If the array is 1D and of shape (n_lipids), the same lipids will be used
            in the registration analysis at every frame. If the array is 2D and of shape
            (n_lipids, n_frames), the boolean value of each lipid at each frame will be
            taken into account. The default is `None`, in which case no filtering is
            performed.
        n_bins : int, optional
            The number of bins to use in each dimension for the two-dimensional density
            calculations. The default is `None`, in which case the number of bins will be given
            by the size of the system in the 'x' dimension rounded up to the nearest integer.
        gaussian_sd : float, optional
            The standard deviation of the circular Gaussian density to convolve with the
            two-dimensional densities. The spreads out the data to better represent the
            size of the lipids. The default is 15.
        """
        
        if not isinstance(leaflets, AssignLeaflets):
            raise ValueError("leaflets must be of type AssignLeaflets")
        
        super(Registration, self).__init__(leaflets.u.trajectory)
        
        self.u = leaflets.u
        self.membrane = leaflets.membrane
        self.leaflets = leaflets.leaflets
        self.start = leaflets.start
        self.stop = leaflets.stop
        self.step = leaflets.step
        
        self.upper_sel = upper_sel if upper_sel is not None else "all"
        self.lower_sel = lower_sel if lower_sel is not None else "all"
        
        if filter_by is not None and np.array(filter_by).ndim not in [1, 2]:
            raise ValueError("'filter_by' must either be a 1D array containing non-changing boolean"
                             "values for each lipid, or a 2D array of shape (n_residues, n_frames)"
                             " containing a boolean value for each lipid at each frame."
                             )

        elif filter_by is not None and len(filter_by) != self.membrane.n_residues:
            raise ValueError("The shape of 'filter_by' must be (n_residues,)")
        
        # determine which lipids to use in the analysis at each frame
        if filter_by is None:
            
            self.filter_by = np.full_like(
                self.leaflets,
                fill_value=True,
                dtype=bool
            )
        elif filter_by.ndim == 1:
            
            self.filter_by = np.full_like(
                self.leaflets,
                fill_value=filter_by[:, np.newaxis],
                dtype=bool
            )
        else:
            self.filter_by = filter_by.astype(bool)
                
        self.n_bins = n_bins
        self.gaussian_sd = gaussian_sd
    
    def _setup_frames(self):
        """
        Pass a Reader object and define the desired iteration pattern
        through the trajectory
        """
        
        self.n_frames = len(range(self.start, self.stop, self.step))
        self.frames = np.zeros(self.n_frames, dtype=int)
        self.times = np.zeros(self.n_frames)
    
    def _prepare(self):
        
        # Output array
        self.registration = np.full(self.n_frames, fill_value=np.NaN)
        
    def _single_frame(self):
        
        # Atoms must be inside the primary unit cell
        self.membrane.residues.atoms.wrap(inplace=True)
        
        # get the bins for the 2d histograms
        x_length = int(np.ceil(self._ts.dimensions)[0])
        if self.n_bins is not None:
            bins = np.linspace(0, x_length, self.n_bins + 1)
        else:
            bins = np.linspace(0, x_length, x_length + 1)
        
        # Upper leaflet 2d histogram
        upper_ordered_res = self.membrane.residues[np.logical_and(
            self.leaflets[:, self._frame_index] == 1,
            self.filter_by[:, self._frame_index],
        )]
        upper_ordered_atoms = upper_ordered_res.atoms.select_atoms(self.upper_sel)
        upper_hist, *_ = np.histogram2d(
            upper_ordered_atoms.positions[:, 0],
            upper_ordered_atoms.positions[:, 1],
            bins=bins
        )
        # convolve with gaussian kernel
        upper_hist = scipy.ndimage.gaussian_filter(
            upper_hist,
            sigma=self.gaussian_sd,
            mode="wrap"
        )
        
        # Find lower ordered atoms
        lower_ordered_res = self.membrane.residues[np.logical_and(
            self.leaflets[:, self._frame_index] == -1,
            self.filter_by[:, self._frame_index],
        )]
        lower_ordered_atoms = lower_ordered_res.atoms.select_atoms(self.lower_sel)
        lower_hist, *_ = np.histogram2d(
            lower_ordered_atoms.positions[:, 0],
            lower_ordered_atoms.positions[:, 1],
            bins=bins
        )
        
        # convolve with gaussian kernel
        lower_hist = scipy.ndimage.gaussian_filter(
            lower_hist,
            sigma=self.gaussian_sd,
            mode="wrap"
        )
        
        # correlation
        correlation = scipy.stats.pearsonr(
            upper_hist.flatten(),
            lower_hist.flatten()
        )
        self.registration[self._frame_index] = correlation[0]
        
    def run(self):
        """Perform the calculation
        """
        
        self._setup_frames()
        self._prepare()
        
        for i, ts in tqdm(enumerate(
                self._trajectory[self.start:self.stop:self.step]),
                total=self.n_frames):
            
            self._frame_index = i
            self._ts = ts
            self.frames[i] = ts.frame
            self.times[i] = ts.time
            
            self._single_frame()
        
        self._conclude()
        return self



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


setup(
    name='lipyphilic',
    version='0.3.0',
    license='LGPL-2.1-or-later',
    description='Analyse MD simulations of lipids with python',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Paul Smith',
    author_email='paul.smith@kcl.ac.uk',
    url='https://github.com/p-j-smith/lipyphilic',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering'
    ],
    project_urls={
        'Documentation': 'https://lipyphilic.readthedocs.io/',
        'Changelog': 'https://lipyphilic.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/p-j-smith/lipyphilic/issues',
    },
    keywords=[
        'lipids', 'molecular dynamics'
    ],
    python_requires='>=3.6,<3.10',
    install_requires=[
        'MDAnalysis>=1.0',
        'freud-analysis==2.4.1',
        'numpy>=1.16,<1.20',  # required for pandas compatibility
        'pandas>=1.1'  # 1.1 required by python=3.6
    ],
    extras_require={
    },
    setup_requires=[
        'pytest-runner',
    ],
    entry_points={
    },
)




import pytest
import numpy as np
import MDAnalysis

from numpy.testing._private.utils import assert_array_equal

from lipyphilic._simple_systems.simple_systems import (
    ONE_CHOL, ONE_CHOL_TRAJ)
from lipyphilic.lib.flip_flop import FlipFlop
 
  
class TestFlipFlop:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(ONE_CHOL, ONE_CHOL_TRAJ)

    @staticmethod
    @pytest.fixture(scope='class')
    def leaflets():
        """Leaflets the ONE_CHOL molecule belongs to at each frame.
        """
        leaflets = np.array(
            [
                -1, -1, -1, -1, -1, -1, 0, 0, 1, 0,
                0, 1, 1, 1, 1, 1, 1, 1, 0, -1,
                -1, -1, -1, -1, -1
            ],
            dtype=np.int8
        )
        #  We need to make sure leaflets has the correct shape
        #  The shape is (n_residues=1, n_frames=25)
        leaflets = leaflets[np.newaxis, :]
        return leaflets

    def test_flip_flop(self, universe, leaflets, lipid_sel="name ROH"):
        
        # 3 flip-flop events with the second one a failure (remains in the upper leaflet)
        flip_flop = FlipFlop(
            universe=universe,
            lipid_sel=lipid_sel,
            leaflets=leaflets,
            frame_cutoff=1,
        )
        flip_flop.run()
        
        reference = {
            'events': [
                [0, 5, 8, 1],
                [0, 8, 11, 1],
                [0, 17, 19, -1]
            ],
            'success': ["Success", "Fail", "Success"]
        }
        
        assert_array_equal(flip_flop.flip_flops, reference['events'])
        assert_array_equal(flip_flop.flip_flop_success, reference['success'])
        
    def test_flip_flop_framecut2(self, universe, leaflets, lipid_sel="name ROH"):
            
        # only two flip-flop events found now, both successful
        flip_flop = FlipFlop(
            universe=universe,
            lipid_sel=lipid_sel,
            leaflets=leaflets,
            frame_cutoff=2,
        )
        flip_flop.run()
        
        reference = {
            'events': [
                [0, 5, 11, 1],
                [0, 17, 19, -1]
            ],
            'success': ["Success", "Success"]
        }
        
        assert_array_equal(flip_flop.flip_flops, reference['events'])
        assert_array_equal(flip_flop.flip_flop_success, reference['success'])

    def test_flip_flop_framecut8(self, universe, leaflets, lipid_sel="name ROH"):
                
        # No flip-flop events found now
        # Cholesterol doesn't remain in opposing lealet for long enough
        flip_flop = FlipFlop(
            universe=universe,
            lipid_sel=lipid_sel,
            leaflets=leaflets,
            frame_cutoff=8,
        )
        flip_flop.run()
        
        reference = {
            'events': np.array([[], [], [], []]).T,
            'success': np.array([])
        }
        
        assert_array_equal(flip_flop.flip_flops, reference['events'])
        assert_array_equal(flip_flop.flip_flop_success, reference['success'])
        
    def test_flip_flop_same_leaflet(self, universe, leaflets, lipid_sel="name ROH"):
                
        flip_flop = FlipFlop(
            universe=universe,
            lipid_sel=lipid_sel,
            leaflets=np.ones_like(leaflets, dtype=np.int8),
        )
        flip_flop.run()
        
        reference = {
            'events': np.array([[], [], [], []]).T,
            'success': np.array([])
        }
        
        assert_array_equal(flip_flop.flip_flops, reference['events'])
        assert_array_equal(flip_flop.flip_flop_success, reference['success'])


class TestAreaPerLipidExceptions:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(ONE_CHOL, ONE_CHOL_TRAJ)

    @staticmethod
    @pytest.fixture(scope='class')
    def leaflets():
        """Leaflets the ONE_CHOL molecule belongs to at each frame.
        """
        leaflets = np.array(
            [
                -1, -1, -1, -1, -1, -1, 0, 0, 1, 0,
                0, 1, 1, 1, 1, 1, 1, 1, 0, -1,
                -1, -1, -1, -1, -1
            ],
            dtype=np.int8
        )
        #  We need to make sure leaflets has he correct shape
        #  The shape is (n_residues=1, n_frames=25)
        leaflets = leaflets[np.newaxis, :]
        return leaflets

    def test_Exceptions(self, universe, leaflets, lipid_sel="name ROH"):
            
        match = ("'leaflets' must be a 2D array of shape \\(n_residues, n_frames\\)"
                 " containing the leaflet id of each lipid at each frame."
                 )
        with pytest.raises(ValueError, match=match):
            FlipFlop(
                universe=universe,
                lipid_sel=lipid_sel,
                leaflets=leaflets[0]  # wrong dimensions
            )
        with pytest.raises(ValueError, match=match):
            FlipFlop(
                universe=universe,
                lipid_sel=lipid_sel,
                leaflets=np.concatenate((leaflets, leaflets))  # wrong number of residues
            )
            
        match = ("The frames to analyse must be identical to those used "
                 "in assigning lipids to leaflets."
                 )
        with pytest.raises(ValueError, match=match):
            flip_flop = FlipFlop(
                universe=universe,
                lipid_sel=lipid_sel,
                leaflets=leaflets[:, ::2]  # not enough frames for the trajectory
            )
            flip_flop.run(
                start=None,
                stop=None,
                step=None
            )
        with pytest.raises(ValueError, match=match):
            flip_flop = FlipFlop(
                universe=universe,
                lipid_sel=lipid_sel,
                leaflets=leaflets
            )
            flip_flop.run(
                start=None,
                stop=None,
                step=2  # different number of frames to what is in 'leaflets'
            )
            
        match = "'frame_cutoff' must be greater than or equal to 1"
        with pytest.raises(ValueError, match=match):
            FlipFlop(
                universe=universe,
                lipid_sel=lipid_sel,
                leaflets=leaflets,
                frame_cutoff=0
            )




import pytest
import numpy as np
import MDAnalysis

from numpy.testing import assert_array_equal

from lipyphilic._simple_systems.simple_systems import (
    HEX_LAT, HEX_LAT_BUMP, HEX_LAT_BUMP_MID_MOL, HEX_LAT_BUMP_MID_ATOM)
from lipyphilic.lib.assign_leaflets import AssignLeaflets
 
 
class TestAssignLeaflets:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)

    kwargs = {
        'lipid_sel': 'name L C'
    }
    
    @pytest.fixture(scope='class')
    def leaflets(self, universe):
        leaflets = AssignLeaflets(universe, **self.kwargs)
        leaflets.run()
        return leaflets

    def test_assign_leaflets(self, leaflets):
    
        reference = {
            'n_residues': 100,
            'n_frames': 1,
            'leaflets_present': [-1, 1]  # all lipids should be assigned to the lower (-1) or upper (1) leaflet
        }
    
        assert leaflets.leaflets.shape == (reference['n_residues'], reference['n_frames'])
        assert_array_equal(np.unique(leaflets.leaflets), reference['leaflets_present'])
    
        # first 50 residues are in the upper leaflet (1)
        # final 50 residues are in the upper leaflet (-1)
        reference = {
            'assigned': np.array([[1]] * 50 + [[-1]] * 50)
        }

        assert_array_equal(leaflets.leaflets, reference['assigned'])


class TestAssignLeafletsExceptions:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT)

    def test_Exceptions(self, universe):
            
        match = "midplane_sel is 'None' and midplane_cutoff "
        with pytest.raises(ValueError, match=match):
            AssignLeaflets(
                universe=universe,
                lipid_sel="name L",
                midplane_sel=None,
                midplane_cutoff=5,
            )
            
        match = "midplane_sel is 'name C' and midplane_cutoff "
        with pytest.raises(ValueError, match=match):
            AssignLeaflets(
                universe=universe,
                lipid_sel="name L C",
                midplane_sel="name C",
                midplane_cutoff=None,
            )
            
        match = "To assign molecules to the midplane, midplane_cutoff must"
        with pytest.raises(ValueError, match=match):
            AssignLeaflets(
                universe=universe,
                lipid_sel="name L C",
                midplane_sel="name C",
                midplane_cutoff=-10,
            )
            
        match = "midplane_sel contains atoms that are not present in molecules selected "
        with pytest.raises(ValueError, match=match):
            AssignLeaflets(
                universe=universe,
                lipid_sel="name L",
                midplane_sel="name C",
                midplane_cutoff=10
            )
        

class TestAssignLeafletsUndulating:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT_BUMP)

    kwargs = {
        'lipid_sel': 'name L C',
        'midplane_sel': 'name C',
        'midplane_cutoff': 6.5,
    }
    
    def test_nbins1(self, universe):
        
        leaflets = AssignLeaflets(universe, n_bins=1, **self.kwargs)
        leaflets.run()
    
        reference = {
            'leaflets_present': [-1, 0, 1],
            'midplane_resnames': ["CHOL"] * 6,  # list or residues incorrectly identified as midplane
        }
    
        assert_array_equal(np.unique(leaflets.leaflets), reference['leaflets_present'])
        assert_array_equal(universe.residues[leaflets.leaflets[:, 0]==0].resnames, reference['midplane_resnames'])  # noqa: E225
        assert 'LIPID' not in universe.residues[leaflets.leaflets[:, 0]==0].resnames  # noqa: E225
        
    def test_nbins4(self, universe):
        
        leaflets = AssignLeaflets(universe, n_bins=4, **self.kwargs)
        leaflets.run()
    
        reference = {
            'leaflets_present': [-1, 1],  # now (correctly) no midplane residues should be found
            'midplane_resnames': [],  # list or residues incorrectly identified as midplane
        }
    
        assert_array_equal(np.unique(leaflets.leaflets), reference['leaflets_present'])
        assert_array_equal(universe.residues[leaflets.leaflets[:, 0]==0].resnames, reference['midplane_resnames'])  # noqa: E225
        

class TestAssignLeafletsUndulatingMidplaneMol:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT_BUMP_MID_MOL)

    kwargs = {
        'lipid_sel': 'name L C',
        'midplane_sel': 'name C',
        'midplane_cutoff': 6.5,
        'n_bins': 4
    }
    
    @pytest.fixture(scope='class')
    def leaflets(self, universe):
        leaflets = AssignLeaflets(universe, **self.kwargs)
        leaflets.run()
        return leaflets
    
    def test_nbins4_midplane(self, universe):
        
        leaflets = AssignLeaflets(universe, **self.kwargs)
        leaflets.run()
    
        reference = {
            'leaflets_present': [-1, 0, 1],
            'midplane_resnames': ["CHOL"],  # list of residues resnames (correctly) identified as midplane
            'midplane_resids': [78]  # resindex of midplane molecules
        }
    
        assert_array_equal(np.unique(leaflets.leaflets), reference['leaflets_present'])
        assert_array_equal(universe.residues[leaflets.leaflets[:, 0] == 0].resnames, reference['midplane_resnames'])
        assert_array_equal(universe.residues[leaflets.leaflets[:, 0] == 0].resids, reference['midplane_resids'])


class TestAssignLeafletsUndulatingMidplaneAtom:
    
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(HEX_LAT_BUMP_MID_ATOM)

    kwargs = {
        'lipid_sel': 'name L C',
        'midplane_sel': 'name C',
        'midplane_cutoff': 6.5,
        'n_bins': 4
    }
    
    @pytest.fixture(scope='class')
    def leaflets(self, universe):
        leaflets = AssignLeaflets(universe, **self.kwargs)
        leaflets.run()
        return leaflets
    
    def test_nbins4_midplane(self, universe):
        
        leaflets = AssignLeaflets(universe, **self.kwargs)
        leaflets.run()
    
        reference = {
            'leaflets_present': [-1, 1],
            'midplane_resnames': [],  # only one of two atoms in CHOL78 is in the midplane, so the molecule is not assigned to the midplane
            'midplane_resids': []
        }
    
        assert_array_equal(np.unique(leaflets.leaflets), reference['leaflets_present'])
        assert_array_equal(universe.residues[leaflets.leaflets[:, 0] == 0].resnames, reference['midplane_resnames'])
        assert_array_equal(universe.residues[leaflets.leaflets[:, 0] == 0].resids, reference['midplane_resids'])
